package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.design.paramils;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.json.JSONObject;

import org.apache.commons.io.FileUtils;

import ca.ubc.cs.beta.hal.algorithms.Algorithm;
import ca.ubc.cs.beta.hal.algorithms.AlgorithmOutputValueTrajectory;
import ca.ubc.cs.beta.hal.algorithms.AlgorithmRunRequest;
import ca.ubc.cs.beta.hal.algorithms.ExternalAlgorithmImplementation;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.TransformedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.metaalgorithms.design.externalconfigurators.AbstractExternalConfiguratorImplementation;
import ca.ubc.cs.beta.hal.algorithms.metaalgorithms.design.externalconfigurators.ConfiguratorRequestHandler;
import ca.ubc.cs.beta.hal.algorithms.parameters.*;
import ca.ubc.cs.beta.hal.algorithms.transformations.DiscretizingTransformation;
import ca.ubc.cs.beta.hal.algorithms.transformations.LinearizingTransformation;
import ca.ubc.cs.beta.hal.algorithms.transformations.Transformation;
import ca.ubc.cs.beta.hal.analysis.Statistics;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.ReadOnlyDataManager;
import ca.ubc.cs.beta.hal.environments.SubrunRunner;
import ca.ubc.cs.beta.hal.problems.*;
import ca.ubc.cs.beta.hal.problems.metaproblems.design.ConfigurationProblem;
import ca.ubc.cs.beta.hal.problems.metrics.*;
import ca.ubc.cs.beta.hal.problems.metrics.Evaluation.Certainty;
import ca.ubc.cs.beta.hal.utils.CombinationIterator;
import ca.ubc.cs.beta.hal.utils.Global;
import ca.ubc.cs.beta.hal.utils.ImmutableJsonSerializable;
import ca.ubc.cs.beta.hal.utils.Misc;

import static ca.ubc.cs.beta.hal.problems.metaproblems.design.ConfigurationProblem.STOP_ON_TIES;
import static ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric.*;


public class ExternalParamILSImplementation extends AbstractExternalConfiguratorImplementation implements ImmutableJsonSerializable {

	public static final String PARAMILS_TAG = "pils";
	public static final String NAME = "HAL_ParamILS", VERSION = "1.1.2";
	
	private static final Logger log = Logger.getLogger(ExternalParamILSImplementation.class.getCanonicalName());
	private static final ParameterSpace SUPP_OPT_SPACE, DFLT_SCEN_SPACE, DFLT_OUT_SPACE, DFLT_CFG_SPACE;
	private static final Pattern INCUMBENT_PARSER = Pattern.compile("\\s*([^=\\s]+)=([^,\\s]+)");

	static {		
		ParameterSpaceBuilder psb = new ParameterSpaceBuilder();
		DFLT_CFG_SPACE = psb.build();
		
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, new RealDomain(0., null, 60.));
		psb.put(Semantics.MAX_SUBRUN_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		SUPP_OPT_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(DISCRETIZATION_FACTOR, new IntegerDomain(1, null, 5));
		psb.put(TRAINING_SET_SIZE, new IntegerDomain(0, null, 2000)); // 0 means don't change the set size
		psb.put(Semantics.SEED, new IntegerDomain(1, Integer.MAX_VALUE, 1));
		psb.put(RESAMPLE_INSTANCES, new BooleanDomain(true));
		psb.put(Semantics.MAX_CPUTIME, new RealDomain(0., null, 24*60.*60.));
		psb.put(Semantics.MAX_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(STOP_ON_TIES, ConfigurationProblem.semanticDomains.get(STOP_ON_TIES));
		psb.put(EVAL_PER_CFG, new IntegerDomain(0, null, 2000));
		psb.setConditional(RESAMPLE_INSTANCES, Misc.asMap(TRAINING_SET_SIZE, (Domain) new IntegerDomain(1, null)));
		DFLT_SCEN_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.DESIGN, Semantics.getDomain(Semantics.DESIGN));
		psb.put(IRUNS, Semantics.getDomain(Semantics.RUNLENGTH));
		psb.put(Semantics.SOLUTION_QUALITY, Semantics.getDomain(Semantics.SOLUTION_QUALITY));
		psb.put(ITIME, new RealDomain(0., null));
		psb.put(ICUTOFF, new RealDomain(0., null));
		DFLT_OUT_SPACE = psb.build();
	}
	
	public ExternalParamILSImplementation() {
		super(NAME, VERSION, 
				Misc.asMap(Semantics.DETERMINISTIC, (Object)false, Semantics.EXPORTABLE, true, Semantics.CUTOFF_AGNOSTIC, true), 
				DFLT_CFG_SPACE, DFLT_SCEN_SPACE, DFLT_OUT_SPACE, SUPP_OPT_SPACE);
	}	
	
	@Override
	public MetaAlgorithmRun getRun(AlgorithmRunRequest request,
			SubrunRunner runner, ReadOnlyDataManager datamanager,
			Statistics stats) {
		return new ExternalParamILSRun(request, runner, datamanager);
	}
	
	@Override
	public JSONObject buildSpec() {
		JSONObject out = super.buildSpec();
		out.put("version", VERSION);
		return out;
	}

	public static ExternalParamILSImplementation fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(ExternalParamILSImplementation.class, spec);
		String ver = o.optString("version");
		if (!VERSION.equals(ver) && ver != null && ver.length() > 0) {			throw new IllegalArgumentException("Record corresponds to different version of ExternalParamILSImplementation plugin; was:" + 
					ver + "; is: " + VERSION );
		}
		return new ExternalParamILSImplementation();
	}

	private static class ExternalParamILSRun extends ExternalConfigurationMetaAlgorithmRun {
	
		private final ParameterSetting scenario, instOptions;
		private final InstanceList instances;
		private final ParameterizedAlgorithm target;
		private final Algorithm configurator;
		private final Random rng;
		private final String metric, objective;
		private final Evaluator metricEvaluator, timeEvaluator;
		private final long seed, targetAlgCutoffSeconds, configuratorRunCutoffTime;
		private final boolean capping;

		private static final String pILSCallJson = "{ " +
			"'command' : 'ruby',"+
			"'deterministic' : false,"+
			"'" + Semantics.CUTOFF_AGNOSTIC + "': false, " +
			"'" + Semantics.EXPORTABLE + "': false, "+
			"'name' : 'Default ParamILS Ruby', " +
			"'tags' : [['" + PARAMILS_TAG + "']]," +
			"'inputFormat' : {"+
			"'callstring' : ['native/paramils/param_ils_2_3_run.rb -numRun $runID$ -scenariofile $scenarioFile$ -maxEvals $maxEvals$ -N $" + EVAL_PER_CFG+"$ -validN 0 -pruning $capping$ [-stopOnTie $stopOnTies$]']},"+
			"'inputs' : {"+
					"'runID' : { "+
							"'semantics':'SEED', 'domain':'Integer(0, 2000000000)'"+
						"},"+
					"'scenarioFile' : { "+
							"'semantics' : '" + Semantics.INSTANCE_FILE + "'"+
						"},"+
					"'"+EVAL_PER_CFG+"' : { "+
							"'domain' : 'Integer(0, inf)', 'default':2000, 'fixed':true"+
							"},"+
					"'maxEvals' : { "+
							"'semantics' : '" + Semantics.MAX_RUNLENGTH + "'" +
						"},"+
					"'capping' : { "+
							"'domain' : 'Integer(0, 1)', 'fixed':true, 'default':1"+
							"},"+
					"'stopOnTies' : { "+
							"'domain' : 'Boolean()', 'fixed':true, 'default':true"+
						"},"+
				"},"+
			"'outputFormat' : { 'stdout' : [ 'New Incumbent: $" + ITIME + "$, $" + IQUALITY + "$ `[$" + IRUNS + "$, $" + ICUTOFF + "$`]. With state $" + IREPORT + "$' ] },"+
			"'outputs' : {"+
				"'" + ITIME + "' : { 'domain' : 'Real[0, inf)' },"+
				"'" + IQUALITY + "' : { 'domain' : 'Real[0, inf)' },"+
				"'" + IRUNS + "' : { 'domain' : 'Integer(0, inf)' },"+
				"'" + ICUTOFF + "' : { 'domain' : 'Real[0, inf)' },"+
				"'" + IREPORT + "' : { 'domain' : 'String(.*$)' }"+
			"} }";
				
		static {
			// make sure PILS has been extracted
			try {
				File basedir = new File(Global.getReferencePoint(), "native/paramils");
				try {
					FileUtils.forceMkdir(basedir);
				} catch (IOException e) {
					if (!basedir.exists()) throw e;
				}
				for (String s: new String[]{"algo_specifics.rb", "global_helper.rb", "param_ils_2_3_run.rb", 
						"param_reader.rb", "stats_ils.rb"}) {
					File target = new File(basedir, s);
					if (!target.exists()) try {
						Misc.copyResourceToFile("/paramils/" + s, target);
					} catch (IOException e) {
						if (!target.exists()) throw e;
					}
				}
			} catch (IOException e) {
				throw new RuntimeException("Error extracting ParamILS files", e);
			}
		}
		
		public ExternalParamILSRun(AlgorithmRunRequest request, SubrunRunner runner, ReadOnlyDataManager dm) {
			super(request, runner, dm);
			this.configurator = ParameterizedAlgorithm.fromSpec(pILSCallJson);
			InstanceMetricMetaProblemInstance inst = (InstanceMetricMetaProblemInstance) request.getProblemInstance();
			PerformanceMetric metric = inst.getMetric();
			scenario = request.getScenario();
			instOptions = inst.getOptions();
			
			String met, obj;
			Boolean cap = true;
			
			// aggregation is either mean or median
			if (metric.getAggregator().equals(ASQ.getAggregator())) {
				met = "mean";
			} else if (metric.getAggregator().equals(MEDR.getAggregator())) {
				met = "median";
				cap = false;
			} else {
				throw new UnsupportedOperationException ("Only mean and median aggregators supported");
			}
			
			Evaluator metEvalCore = metric.getEvaluator();
			
			while (metEvalCore instanceof EmbeddedEvaluator) {
				metEvalCore = ((EmbeddedEvaluator)metEvalCore).getCore();
			}
			
			metricEvaluator = metric.getEvaluator();
			
			
			
			// evaluator is either P_R1, P_R10, or SQ 
			if (Misc.asSet(PAR1.getEvaluator(), ReportedPAR1.getEvaluator(), MAXPAR1.getEvaluator()).contains(metEvalCore)) {
				obj = "runtime";
				timeEvaluator = metricEvaluator;
			} else if (Misc.asSet(PAR10.getEvaluator(), ReportedPAR10.getEvaluator(), MAXPAR10.getEvaluator()).contains(metEvalCore)) {
				if (!"mean".equals(met)) {
					throw new UnsupportedOperationException("Penalized-10 runtime only supported with mean aggregator");
				}
				obj = "runtime";
				met = "mean10";
				timeEvaluator = metricEvaluator;
			} else {
				obj = "qual";
				cap = false;
				timeEvaluator = maxCpu;
			} 
			this.metric = met;
			this.objective = obj;
			this.capping = cap;

			seed = ((Number)scenario.get(Semantics.SEED)).longValue();
			rng = Global.getRandom(seed);

			instances = setupInstanceList(request);

			ParameterizedAlgorithm targ = (ParameterizedAlgorithm) inst.getAlgorithms();
			// make sure target alg domains are linearized
			if (!LinearizingTransformation.isLinear(targ)) {
				Transformation t = new LinearizingTransformation();
				log.info("Linearizing all numerical parameters with nonlinear scale");
				targ = t.transform(targ);
			}
			// make sure target alg is discretized
			if (!DiscretizingTransformation.isFiniteDiscrete(targ)) {
				int discFactor = ((Number)scenario.get(DISCRETIZATION_FACTOR)).intValue();
				Transformation t = new DiscretizingTransformation(discFactor);
				targ = t.transform(targ);
				log.warning("Target algorithm has continuous parameters but ParamILS requires discrete domains.  Discretizing using " + t + "\n" +
						targ.getConfigurationSpace());
			}
			target = targ;
			log.info("Target algorithm: " + target);

			Number targcutoff = (Number)instOptions.get(Semantics.MAX_SUBRUN_CPUTIME);
			if (targcutoff == null) {
				if (target.getScenarioValue(Semantics.MAX_CPUTIME) != null) {
					targcutoff = (Number) target.getScenarioValue(Semantics.MAX_CPUTIME);
					log.warning("As no MAX_SUBRUN_CPUTIME option set in meta-instance, using current value of MAX_CPUTIME for target algorithm: " + targcutoff);
				} else {
					targcutoff = (Number) SUPP_OPT_SPACE.getDefaultValue(Semantics.MAX_SUBRUN_CPUTIME);
					log.warning("As no MAX_SUBRUN_CPUTIME option set in meta-instance, using default: " + targcutoff);
				}
			}
			Number cfgcutoff = (Number)scenario.get(Semantics.MAX_CPUTIME);
			
			int targetAlgCutoffSeconds = (int) Math.ceil(targcutoff.doubleValue());
			int configuratorRunCutoffTime = (int) Math.ceil(cfgcutoff.doubleValue());
			if (targcutoff.doubleValue() != (double) targetAlgCutoffSeconds) {
				log.info("Using " + targetAlgCutoffSeconds + " as target algorithm max cputime instead of requested " + targcutoff);
			}
			if (cfgcutoff.doubleValue() != (double) configuratorRunCutoffTime) {
				log.info("Using " + configuratorRunCutoffTime + " as configurator max cputime instead of requested " + cfgcutoff);
			}
			this.targetAlgCutoffSeconds = targetAlgCutoffSeconds;
			this.configuratorRunCutoffTime = configuratorRunCutoffTime;
		}

		@Override
		protected AlgorithmRunRequest getTargetRunRequest(String instance,
				long seed, double runtimeCutoff, long runlengthCutoff,
				ParameterSetting configuration) {
			
			int instnum = Integer.valueOf(new File(instance).getName());
			synchronized(target) {
				ParameterSettingBuilder newScenario = new ParameterSettingBuilder(target.getScenario()),
					newConfig = new ParameterSettingBuilder(target.getConfiguration());
				for (Entry<String, Object> e: configuration.entrySet()) {
					newConfig.put(e.getKey(), unescapeSpace(e.getValue()));
				}
				newScenario.put(Semantics.SEED, seed);
				newScenario.put(Semantics.MAX_CPUTIME, runtimeCutoff);
				newScenario.put(Semantics.MAX_RUNLENGTH, runlengthCutoff);
				target.setProblemInstance(instances.get(instnum));
				target.updateScenario(newScenario.build(), true);
				target.updateConfiguration(newConfig.build(), true);
				AlgorithmRunRequest request = target.getAlgorithmRunRequest();
				request.setDistributedExecutionOK(false);
				return request;
			}
		}

		private static final String SPACE_HOLDER = "_HalPilsSpace_";
		
		private static final Object escapeSpace(Object o) {
			if (o instanceof String) {
				return ((String)o).replace(" ", SPACE_HOLDER);
			}
			return o;
		}
		
		private static final Object unescapeSpace(Object o) {
			if (o instanceof String) {
				return ((String)o).replace(SPACE_HOLDER, " ");
			}
			return o;
		}
		
		@SuppressWarnings("unchecked")
		@Override
		protected AlgorithmRunRequest getConfiguratorRunRequest() {
			
			File scratchDir = new File(Global.getReferencePoint(), "PILS_scratch");
			try {
				FileUtils.forceMkdir(scratchDir);
			} catch (IOException e) {
				if (!scratchDir.exists()) throw new RuntimeException("Errors creating scratch space", e);
			}
			
			long id = this.getId();
			File pILSOutputDirectory = new File(scratchDir, "paramILS-run"+id+"_" +System.currentTimeMillis());		
			try {
				FileUtils.forceMkdir(pILSOutputDirectory);
			} catch (IOException e) {
				throw new RuntimeException("Errors creating paramils temp directory", e);
			}
			
			// write parameter file
			File parameterFile = new File(pILSOutputDirectory, "parameters.txt");
			try {
				PrintStream parameterWriter = new PrintStream(parameterFile);
				
				Map<String, Object[]> discretisedDomains = new HashMap<String, Object[]>();
				
				Set<String> targetInputs = target.getConfigurationSpace().keySet();
				for (String key : targetInputs) {
					Domain d0 = target.getConfigurationDomain(key);
					parameterWriter.print(key);
					parameterWriter.print(" {");

					DiscreteDomain<?> d = (DiscreteDomain<?>)d0;
					Iterator<?> it = d.iterator();
					boolean wroteDefault = false;
					boolean first = true;

					Object[] disc = new Object[(int)d.size()];
					int discIndex = 0;

					while (it.hasNext()) {
						Object o = it.next();

						disc[discIndex] = o;
						discIndex++;

						if (!first) {
							parameterWriter.print(", ");
						} else {
							first = false;
						}
						parameterWriter.print(escapeSpace(o));

						if (o.equals(d.getDefaultValue())) {
							wroteDefault = true;
						}
					}
					discretisedDomains.put(key, disc);
					if (!wroteDefault) {
						parameterWriter.print(", "+escapeSpace(d.getDefaultValue()));
					}
					parameterWriter.print("} [");
					parameterWriter.print(escapeSpace(d0.getDefaultValue()));
					parameterWriter.println("]");
				}
				
				// conditional parameters, if any. Have to write the string regardless because PILS has a terrible parser.
				parameterWriter.println("");
				parameterWriter.println("Conditionals:");
				Map<String, List<Map<String, Domain>>> conditionals = target.getConfigurationSpace().getConditionalConfigs();
				if (conditionals.keySet().size() > 0) {	
					for (String key : conditionals.keySet()) {
						Collection<Map<String, Domain>> disjunction = conditionals.get(key);
						// ParamILS can only support a single set of conjunctive conditions on a parameter.
						if (disjunction.size() == 1) {
							// each one printed out separately.
							Map<String, Domain> conjuncts = disjunction.iterator().next();
							for (String conKey : conjuncts.keySet()) {
								Domain d = conjuncts.get(conKey);
								parameterWriter.print(key+" | "+conKey+" in {");
								boolean first = true;
								for (Object o : discretisedDomains.get(conKey)) {
									if (d.contains(o)) {
										if (!first) {
											parameterWriter.print(",");
										} else {
											first = false;
										}
										parameterWriter.print(escapeSpace(o));
									}
								}
								parameterWriter.println("}");
							}
						} else {
							log.warning("ParamILS doesn't support disjunctive conditions on a single variable; removing conditions on " + key);
//							throw new IllegalArgumentException("ParamILS doesn't support disjunctive conditions on a single variable! "+key);
						}
					}
				}
				parameterWriter.println();
				
				// prohibited configurations, if any
				Collection<Map<String, Domain>> prohibited = target.getConfigurationSpace().getProhibitedConfigs();
				if (prohibited.size() > 0) {
					parameterWriter.println();
					parameterWriter.println("Forbidden:");
					for (Map<String, Domain> disjunct : prohibited) {
						// This corresponds to a set of prohibited configurations.
						// Build all the filtered domains, and use CombinationIterator to do the heavy lifting.
						ArrayList<ArrayList<Object>> filteredDomains = new ArrayList<ArrayList<Object>>();
						ArrayList<String> keys = new ArrayList<String>();
						
						for (String s : disjunct.keySet()) {
							Domain d = disjunct.get(s);
							ArrayList<Object> f = new ArrayList<Object>();
							for (Object obj : discretisedDomains.get(s)) {
								if (d.contains(obj)) {
									f.add(obj);
								}
							}
							keys.add(s);
							filteredDomains.add(f);
						}
						CombinationIterator<Object> iter = new CombinationIterator<Object>(filteredDomains);
						for (List<Object> combination : iter) {
							parameterWriter.print("{");
							Iterator<Object> it1 = combination.iterator();
							Iterator<String> names = keys.iterator();
							boolean first = true;
							while (it1.hasNext() && names.hasNext()) {
								Object val = it1.next();
								String name = names.next();
								if (!first) {
									parameterWriter.print(",");
								} else {
									first = false;
								}
								parameterWriter.print(name+"="+escapeSpace(val));
							}
							parameterWriter.println("}");
						}
						parameterWriter.println();
					}
				}
				parameterWriter.close();
			} catch (IOException e) {
				throw new RuntimeException("Error writing ParamILS parameter file.", e);
			}
			
			// write training instance+seed file
			SampleableDomain<? extends Number> seedDom = (SampleableDomain<? extends Number>) (
					target.hasScenarioVariable(Semantics.SEED) ? 
					target.getScenarioDomain(Semantics.SEED) : Semantics.getDomain(Semantics.SEED));
			File instanceSeedFile = new File(pILSOutputDirectory, "training.txt");
			File instanceSeedNameFile = new File(pILSOutputDirectory, "trainingNames.txt");
			try {
				PrintStream instanceWriter = new PrintStream(instanceSeedFile);	
				PrintStream instanceNameWriter = new PrintStream(instanceSeedNameFile);				
				Map<ProblemInstance, Random> seedGenerators = new HashMap<ProblemInstance, Random>();
				for (int i=0; i<instances.size(); i++) {
					ProblemInstance pi = instances.get(i);
					if (!seedGenerators.containsKey(pi)) seedGenerators.put(pi, Global.getRandom(rng.nextInt()));
					Random seedGenerator = seedGenerators.get(pi);
					long seed = target.isDeterministic() ? -1 : seedDom.getSample(seedGenerator).longValue();
					instanceWriter.println(seed+" "+i);
					instanceNameWriter.println(seed+" "+ 
							(target.getImplementation() instanceof ExternalAlgorithmImplementation ?
									((ExternalAlgorithmImplementation)target.getImplementation()).resolvePath(pi.getFeatureValue(Semantics.INSTANCE_FILE)) :
									pi.getFeatureValue(Semantics.INSTANCE_FILE)));
				}				
				instanceWriter.close();
				instanceNameWriter.close();
			} catch (IOException e) {
				throw new RuntimeException("Error writing ParamILS instance file.", e);
			}
			
			// write scenario file
			File scenarioFile = new File(pILSOutputDirectory, "scenario.txt");
			File rubyFile = new File(pILSOutputDirectory, "rubyWrapper.rb");
			File outDir = new File(pILSOutputDirectory, "paramils_out");
			try {
				PrintStream scenarioWriter = new PrintStream(scenarioFile);
				scenarioWriter.println("algo = ruby " + rubyFile.getPath());
				scenarioWriter.println("execdir = ./");
				scenarioWriter.println("deterministic = "+(target.isDeterministic() ? "1" : "0"));
				scenarioWriter.println("run_obj = " + objective);
				scenarioWriter.println("overall_obj = " + metric);
				scenarioWriter.println("cutoff_time = "+targetAlgCutoffSeconds);
				scenarioWriter.println("tunerTimeout = "+configuratorRunCutoffTime);
				scenarioWriter.println("paramfile = "+parameterFile.getPath());
				scenarioWriter.println("outdir = " + outDir.getPath());
				scenarioWriter.println("instance_seed_file = "+instanceSeedFile.getPath());
				scenarioWriter.println("test_instance_seed_file = "+instanceSeedFile.getPath());
				scenarioWriter.close();
			} catch (IOException e) {
				throw new RuntimeException("Error writing ParamILS scenario", e);
			}
			
			try {
				String wrapperText = Misc.readResourceToString("/"+getClass().getPackage().getName().replace(".", "/") + "/rubyWrapperTemplate.rb");
				String modifiedWrapper = wrapperText.replaceAll("\\$hostName\\$", "\""+rpcUrl.getHost()+"\"");
				modifiedWrapper = modifiedWrapper.replaceAll("\\$hostPort\\$", String.valueOf(rpcUrl.getPort()));
				modifiedWrapper = modifiedWrapper.replaceAll("\\$experimentID\\$", String.valueOf(getId()));
				modifiedWrapper = modifiedWrapper.replaceAll("\\$rpcHandler\\$", "\""+ConfiguratorRequestHandler.class.getCanonicalName() + ".\"");
				modifiedWrapper = modifiedWrapper.replaceAll("\\$requestRunCommand\\$", "\"doRun\"");
				FileUtils.writeStringToFile(rubyFile, modifiedWrapper);
			} catch (IOException e) {
				throw new RuntimeException("Error copying the wrapper template.", e);
			}
			
			ProblemInstance pilsInst;
			try {
				pilsInst = new FileProblemInstance(scenarioFile);
				pilsInst.addTags(PARAMILS_TAG);
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
			synchronized(configurator) {
				
				Number seedNum = seed;
				if (!configurator.getScenarioDomain(Semantics.SEED).contains(seedNum)) {
					// we can't pass the provided seed straight through in this case
					Random r = Global.getRandom(seed);
					seedNum = ((SampleableDomain<? extends Number>)configurator.getScenarioDomain(Semantics.SEED)).getSample(r);
				}

				BigInteger maxevals = BigInteger.valueOf(((Number)scenario.get(Semantics.MAX_RUNLENGTH)).longValue()); 
				BigInteger maxuniqueruns = target.getConfigurationSpace().numDiscreteConfigurations().multiply(BigInteger.valueOf(instances.size()));
				maxevals = maxuniqueruns.min(maxevals);
				maxevals = maxevals.min(BigInteger.valueOf(Long.MAX_VALUE));
				int N = ((Number)scenario.get(EVAL_PER_CFG)).intValue();
				N = (int) Math.min(N, instances.size());
				configurator.setScenarioValue(Semantics.SEED, seedNum);
				configurator.setScenarioValue(Semantics.MAX_CPUTIME, scenario.get(Semantics.MAX_CPUTIME));
				configurator.setScenarioValue(Semantics.MAX_RUNLENGTH, maxevals.longValue());
				configurator.setScenarioValue(EVAL_PER_CFG, N);
				if (!capping) {
					configurator.setScenarioValue("capping", 0);
				} else {
					configurator.setScenarioValue("capping", 1);
				}
				configurator.setProblemInstance(pilsInst);	
				if (scenario.containsKey(STOP_ON_TIES)) {
					configurator.setScenarioValue(STOP_ON_TIES, scenario.get(STOP_ON_TIES));
				} else {
					configurator.setScenarioValue(STOP_ON_TIES, DFLT_SCEN_SPACE.getDefaultValue(STOP_ON_TIES));
				}
				System.out.println(((ExternalAlgorithmImplementation)configurator.getImplementation()).getCommandString(configurator.getAlgorithmRunRequest()));
				return configurator.getAlgorithmRunRequest();
			}
		}

		@Override
		protected String getConfiguratorSpecificResultString(AlgorithmRun run,
				String instance, long seed, double runtimeCutoff,
				long runlengthCutoff) {
			
			Evaluation metEval = metricEvaluator.getEvaluation(run), timeEval = timeEvaluator.getEvaluation(run);
			
			StringBuffer result = new StringBuffer();
			result.append("Result for ParamILS: ");
			if (RunStatus.finishedWithUnexpectedError(run.getStatus())) {
				result.append("CRASHED, ");
			} else if (metEval.getCertainty() != Certainty.CERTAIN) {
				result.append("TIMEOUT, ");
			} else {
				result.append("SAT, ");
			}

			double runtime;
			try {
				// not using the metric score directly because we can't report anything but true runtime
				// the way we get that, is by looking at the cap; we've made sure earlier that
				// the runtime evaluator is indeed a runtime evaluator, which we know will produce this
				runtime = Evaluation.getCapValue(timeEval, Semantics.MAX_CPUTIME).doubleValue();
			} catch (NoSuchElementException e) {
				// we NEED a runtime, so if the evaluator fails, use HAL's measurement as a last resort.
				log.log(Level.WARNING, 
						"Evaluator " + timeEvaluator + " failed to determine runtime for run " + run + "; using HAL-measured CPU time", e);
				runtime = run.getMeasuredCpuTime();
			}
			
			result.append(runtime);
			result.append(", ");
			
			
			Number d = null;
			if (run.hasOutputVariable(Semantics.RUNLENGTH)) {
				d = (Number) run.getLastOutputValueOnly(Semantics.RUNLENGTH);
			}
			result.append((d == null ? runlengthCutoff : d));
			result.append(", ");

			d = metEval.getValue();
			if (this.objective != "qual" && run.hasOutputVariable(Semantics.SOLUTION_QUALITY)) {
				d = (Number) run.getLastOutputValueOnly(Semantics.SOLUTION_QUALITY);
			} 
			d = d == null ? runtime : d;
			result.append(d);
			result.append(", ");
			result.append(seed);
			
			System.err.println(result);
			
			return result.toString();
		}

		// variables used when parsing PILS outputs and deciding whether to produce corresponding HAL output.
		private volatile int lastOutputSize = 0;
		private volatile String lastReport = "";
		@Override
		protected void reportConfiguratorRun(AlgorithmRun r) {
			// output visitor:  every time the underlying algorithm produces an output
			// we might want to produce an output for HAL	
			AlgorithmRunVisitor outputVisitor = new AlgorithmRunVisitor() {
				@Override
				public synchronized void visit(AlgorithmRun r) {
					ParameterSetting incumbent = null;
					AlgorithmOutputValueTrajectory ireports = r.getOutput(IREPORT);
					if (ireports == null) {
						log.warning("Recieved parse request but no incumbents reported");
						return;
					}
					int newsize = ireports.size();
					if (newsize <= lastOutputSize) return; // no new incumbent outputs
					lastOutputSize = newsize;
							
					Map<String, Object> toupdate = new HashMap<String, Object>();
					// parse and add any new incumbent
					Object	incumbentReport = ireports.getLast().getValue(),
							incumbentTime = r.getLastOutputValueOnly(ITIME),
							incumbentQuality = r.getLastOutputValueOnly(IQUALITY),
							incumbentRun = r.getLastOutputValueOnly(IRUNS),
							incumbentCutoff = r.getLastOutputValueOnly(ICUTOFF);
					if (incumbentReport != null && !incumbentReport.equals(lastReport)) {
						incumbent = parseIncumbent((String)incumbentReport);
						lastReport = (String) incumbentReport;
						synchronized(target) {
							target.updateConfiguration(incumbent);
							ParameterizedAlgorithm trg = target;
							if (trg instanceof TransformedAlgorithm) {
								trg = ((TransformedAlgorithm)trg).getCore();
							}
							toupdate.put(Semantics.DESIGN, trg.getParameterlessAlgorithm());
						}
					}
					toupdate.put(ITIME, incumbentTime);
					toupdate.put(Semantics.SOLUTION_QUALITY, incumbentQuality);
					toupdate.put(Semantics.RUNLENGTH, incumbentRun);
					toupdate.put(ICUTOFF, incumbentCutoff);
					ExternalParamILSRun.this.updateOutput(toupdate);
				}
			};
			r.registerOutputVisitor(outputVisitor);
		}
		
		/***
		 * Method to read a paramILS output string and create a corresponding ParameterSetting
		 * @param incumbentReport
		 * @return
		 */
		private ParameterSetting parseIncumbent(String incumbentReport) {
			ParameterSpace cspace = target.getConfigurationSpace();
			if (incumbentReport == null || incumbentReport.length() == 0) return null;
			Matcher m = INCUMBENT_PARSER.matcher(incumbentReport);
			ParameterSettingBuilder newconfig = new ParameterSettingBuilder(cspace.getDefaults());
			int end = 0;
			while (m.find()) {
				String var = m.group(1);
				Object val = cspace.get(var).cast(unescapeSpace(m.group(2)));
				newconfig.put(var, val);
				end = m.end();
			}
			if (incumbentReport.substring(end).trim().length() > 0) {
				throw new RuntimeException("Incomplete incumbent report: " + incumbentReport);
			}
			return cspace.reduceSetting(newconfig, true).build();
		}
	}

	@Override
	public Set<Set<String>> getRequiredTags() {
		//return DFLT_REQ_TAGS;
		
		Set<Set<String>> tags = super.getRequiredTags();
		Set<String> moreTags = new TreeSet<String>();
		moreTags.addAll(Misc.asSet(MetaProblemInstance.TAG));
		tags.add(moreTags);
		
		return tags;
	}
}
