package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis;

import static ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis.AnalysisSemantics.*;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.logging.Logger;

import net.sf.json.JSONObject;

import org.apache.commons.math.stat.descriptive.DescriptiveStatistics;

import ca.ubc.cs.beta.hal.algorithms.Algorithm;
import ca.ubc.cs.beta.hal.algorithms.AlgorithmRunRequest;
import ca.ubc.cs.beta.hal.algorithms.MetaAlgorithmImplementation;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithmDistribution;
import ca.ubc.cs.beta.hal.algorithms.parameters.BooleanDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.IntegerDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ListDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ObjectClassDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpace;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpaceBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.RealDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.SampleableDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.analysis.Plot;
import ca.ubc.cs.beta.hal.analysis.Plot.Layer;
import ca.ubc.cs.beta.hal.analysis.Statistics;
import ca.ubc.cs.beta.hal.analysis.Statistics.Alternative;
import ca.ubc.cs.beta.hal.analysis.TestResult;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.SubrunRunner;
import ca.ubc.cs.beta.hal.problems.InstanceDistribution;
import ca.ubc.cs.beta.hal.problems.InstanceMetricMetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.MetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;
import ca.ubc.cs.beta.hal.utils.ImmutableJsonSerializable;
import ca.ubc.cs.beta.hal.utils.Misc;

public class KWayComparison extends	MetaAlgorithmImplementation implements ImmutableJsonSerializable {
	
	private static final Logger log = Logger.getLogger(ComprehensivePairwiseComparison.class.getCanonicalName());
	private static final String VERSION = "1.1.2";
	public static final String FRIEDMAN_Q = "Friedman Q", FRIEDMAN_P = "Friedman p",
			WILCOXON_WS = "Wilcoxon W's", WILCOXON_PS = "Wilcoxon p's",
			SIGNIFICANCE = "Significance Level", WILCOXON_WINNERS = "WilcoxonWinners";
	private static final ParameterSpace DFLT_OUT_SPACE, DFLT_SCN_SPACE, DFLT_CFG_SPACE, SUPP_OPT_SPACE;
	static {
		ParameterSpaceBuilder psb = new ParameterSpaceBuilder();
		psb.put(Semantics.RUNLENGTH, Semantics.getDomain(Semantics.RUNLENGTH));
		psb.put(SCD_OVERLAY, new ObjectClassDomain(Plot.class));
		psb.put(SOLUTION_QUALITIES, new ListDomain(new RealDomain()));
		psb.put(MEANS, psb.get(SOLUTION_QUALITIES));
		psb.put(STDDEVS, psb.get(SOLUTION_QUALITIES));
		psb.put(Q90S, psb.get(SOLUTION_QUALITIES));
		psb.put(Q75S, psb.get(SOLUTION_QUALITIES));
		psb.put(Q50S, psb.get(SOLUTION_QUALITIES));
		psb.put(Q25S, psb.get(SOLUTION_QUALITIES));
		psb.put(Q10S, psb.get(SOLUTION_QUALITIES));
		psb.put(RUNLENGTHS, new ListDomain(psb.get(Semantics.RUNLENGTH)));
		psb.put(SCDS, new ListDomain(new ObjectClassDomain(Plot.class)));
		psb.put(FRIEDMAN_Q, new RealDomain(0., null));
		psb.put(FRIEDMAN_P, new RealDomain(0., 1.));
		psb.put(WILCOXON_WS, new ListDomain(new IntegerDomain(0, null)));
		psb.put(WILCOXON_PS, new ListDomain(new RealDomain(0., 1.)));
		psb.put(WILCOXON_WINNERS, new ListDomain(new BooleanDomain()));
		DFLT_OUT_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		DFLT_CFG_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.SEED, Semantics.getDomain(Semantics.SEED));
		psb.put(Semantics.MAX_CPUTIME, new RealDomain(0.0, null, 1e300));
		psb.put(Semantics.MAX_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(MIN_OUTPUT_INTERVAL, new RealDomain(0., null, 60.0));
		psb.put(COMBINE_IDENTICAL_INSTANCES, new BooleanDomain(false));
		psb.put(MAX_PARALLEL_RUNS, new IntegerDomain(1, null, 100));
		psb.put(SIGNIFICANCE, new RealDomain(0., 1.));
		DFLT_SCN_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, new RealDomain(0.0, null, 1e300));
		psb.put(Semantics.MAX_SUBRUN_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(Semantics.RUNS_PER_INSTANCE, new IntegerDomain(1L, null, 1L));
		SUPP_OPT_SPACE = psb.build();
	}
	
	private static final Map<String, String> variables = Misc.asMap(
			SCD_OVERLAY, "solution cost distributions",
			SCATTER, "correlation plot",
			SCDS, "per-algorithm solution cost distributions",
			SOLUTION_QUALITIES, "per-algorithm solution quality per provided metrics",
			MEANS, "per-algorithm mean performances",
			STDDEVS, "per-algorithm performance standard deviations",
			Q90S, "per-algorithm 90th performance percentiles",
			Q75S, "per-algorithm 75th performance percentiles",
			Q50S, "per-algorithm median performances",
			Q25S, "per-algorithm 25th performance percentiles",
			Q10S, "per-algorithm 10th performance percentiles",
			MIN_OUTPUT_INTERVAL, "minimum logged output interval (s)",
			COMBINE_IDENTICAL_INSTANCES, "collapse instances with identical hash",
			Semantics.MAX_RUNLENGTH, "maximum number of target algorithm runs",
			MAX_PARALLEL_RUNS, "maximum runs to launch in parallel");
	
	public String describeSemantics(String semantics) {
		String out = variables.get(semantics);
		return out == null ? Semantics.describe(semantics) : out;
	}
	public Collection<String> getDefinedSemantics() {
		return variables.keySet();
	}

	protected KWayComparison(String name, String version,
			ParameterSpace defaultConfigurationSpace,
			ParameterSpace defaultScenarioSpace,
			ParameterSpace defaultOutputSpace,
			ParameterSpace supportedInstanceOptionSpace,
			Object... propertyvals) {
		super(name, version, defaultConfigurationSpace, defaultScenarioSpace, defaultOutputSpace,
				supportedInstanceOptionSpace, propertyvals);
	}
	
	public KWayComparison() {
		super(KWayComparison.class.getSimpleName(), VERSION,
				DFLT_CFG_SPACE, DFLT_SCN_SPACE, DFLT_OUT_SPACE, SUPP_OPT_SPACE,
				Semantics.DETERMINISTIC, false);
	}
	
	@SuppressWarnings("unchecked")
	public Set<Set<String>> getRequiredTags() {
		return Misc.asSet(Misc.asSet(MetaProblemInstance.TAG, InstanceMetricMetaProblemInstance.TAG));
	}

	@Override
	public MetaAlgorithmRun getRun(AlgorithmRunRequest request,
			SubrunRunner runner, Statistics stats) {
		if (!request.getImplementation().equals(this)) {
			throw new IllegalArgumentException("Request does not correspond to this meta-algorithm implementation");
		}
		return new KWayComparisonRun(request, runner, stats);
	}
	
	@Override
	public JSONObject buildSpec() {
		JSONObject out = super.buildSpec();
		//out.put("version", VERSION);
		return out;
	}
	
	@Override
	public String getVersion() {
		return VERSION;
	}

	public static KWayComparison fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(KWayComparison.class, spec);
		String ver = o.optString("version");
		if (!VERSION.equals(ver) && ver != null && ver.length() > 0) {
			throw new IllegalArgumentException("Record corresponds to different version of KWayComparison plugin; was:" +
					ver + "; is: " + VERSION );
		}
		return new KWayComparison();
	}
	
	protected static class KWayComparisonRun extends MetaAlgorithmRun {

		protected final ParameterlessAlgorithmDistribution targets;
		protected final InstanceDistribution distribution;
		protected final PerformanceMetric metric;
		private final Statistics statistics;
		private final SubrunRunner runner;
		private final Random rng;
		protected final Map<ParameterlessAlgorithm, Map<Long, List<Double>>> data = new HashMap<ParameterlessAlgorithm, Map<Long, List<Double>>>();
		private final Map<AlgorithmRun, Long> instids = new HashMap<AlgorithmRun, Long>();
		private final InstanceMetricMetaProblemInstance instance;
		private final Double maxsubruntime, outinterval, significance;
		private final Long maxruns, runsperinst, maxsubrunlength, maxParallel;
		private volatile double maxtime, freetime, lasttime=0.;
		private volatile long runsleft, inprog, done=0L;
		protected volatile double extraCpuTime = 0, myVisitorTime = 0;
		private final boolean combine;
		private final SampleableDomain<? extends Number> seedDomain;
		
		@SuppressWarnings("unchecked")
		public KWayComparisonRun(
				AlgorithmRunRequest request, SubrunRunner runner,
				Statistics stats) {
			super(request, runner);
			this.runner = runner;
			combine = (Boolean) request.getScenarioValue(COMBINE_IDENTICAL_INSTANCES);
			ProblemInstance inst0 = request.getProblemInstance();
			assert inst0 instanceof InstanceMetricMetaProblemInstance;
			
			instance = (InstanceMetricMetaProblemInstance) inst0;
			assert instance.getAlgorithms().size().compareTo(BigInteger.ZERO) > 0;
			assert instance.getInstanceDistribution().size() > 0;
			assert instance.getMetric() != null;
			
			this.statistics = stats;
			maxParallel = ((Number)request.getScenarioValue(MAX_PARALLEL_RUNS)).longValue();
			
			metric = instance.getMetric();
			distribution = instance.getInstanceDistribution();
			targets = instance.getAlgorithms();

			if (targets instanceof ParameterizedAlgorithm) {
				throw new IllegalArgumentException("Can't operate on a single ParameterizedAlgorithm");
			}
			
			SampleableDomain<? extends Number> d = (SampleableDomain<? extends Number>) Semantics.getDomain(Semantics.SEED);
			for (Algorithm a: targets) {
				if (a.hasScenarioVariable(Semantics.SEED)) {
					d = (SampleableDomain<? extends Number>) d.intersection(a.getScenarioDomain(Semantics.SEED));
				}
			}
			seedDomain = d;
			
			Number n = (Number) request.getScenarioValue(Semantics.SEED);
			rng = Global.getRandom(n == null ? System.currentTimeMillis() : n.longValue());
			
			n = (Number) request.getScenarioValue(SIGNIFICANCE);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(SIGNIFICANCE).getDefaultValue();
			significance = n.doubleValue();
			
			n = (Number) request.getScenarioValue(Semantics.MAX_CPUTIME);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(Semantics.MAX_CPUTIME).getDefaultValue();
			maxtime = n.doubleValue();
			freetime = maxtime;
			
			n = (Number) request.getScenarioValue(MIN_OUTPUT_INTERVAL);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(MIN_OUTPUT_INTERVAL).getDefaultValue();
			outinterval = n.doubleValue();
			
			n = (Number) instance.getOption(Semantics.RUNS_PER_INSTANCE);
			if (n == null) n = (Number) SUPP_OPT_SPACE.get(Semantics.RUNS_PER_INSTANCE).getDefaultValue();
			runsperinst = n.longValue();
			
			n = (Number) request.getScenarioValue(Semantics.MAX_RUNLENGTH);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(Semantics.MAX_RUNLENGTH).getDefaultValue();
			if (distribution.size() > 0 && runsperinst > 0) {
				n = Math.min(n.longValue(), distribution.size()*runsperinst*targets.size().intValue());
			}
			runsleft = n.longValue();
			maxruns = runsleft;
						
			n = (Number) instance.getOption(Semantics.MAX_SUBRUN_CPUTIME);
			if (n != null && n.doubleValue() < Double.MAX_VALUE) {
				maxsubruntime = Math.min(n.doubleValue(), maxtime);
			} else {
				maxsubruntime = Double.MAX_VALUE;
			}
			n = (Number) instance.getOption(Semantics.MAX_SUBRUN_RUNLENGTH);
			if (n != null && n.longValue() < Long.MAX_VALUE) {
				maxsubrunlength = n.longValue();
			} else {
				maxsubrunlength = Long.MAX_VALUE;
			}
		}

		private static final boolean matches(AlgorithmRunRequest rrq, ParameterlessAlgorithm a) {
			return rrq.getImplementation().equals(a.getImplementation()) && rrq.getConfiguration().equals(a.getConfiguration());
		}
		
		private final ParameterlessAlgorithm getAlg(AlgorithmRunRequest rrq) {
			for (ParameterlessAlgorithm a: targets) if (matches(rrq, a)) return a;
			throw new IllegalArgumentException("no such element " + rrq);
		}
		
		private final Set<String> retries = new HashSet<String>();
		private class ComparisonSubrunCompletionVisitor extends AlgorithmRunVisitor {
			@Override
			public void reportRunTime(double time) {
				super.reportRunTime(time);
				myVisitorTime += time;
			}
			@Override
			public void visit(AlgorithmRun r) {
				Double time = getMeasuredCpuTime();
				ParameterlessAlgorithm impl = getAlg(r.getAlgorithmRunRequest());
				synchronized(data) {
					try {
						Long instid = instids.remove(r);
						if (instid == null) throw new RuntimeException("Unexpected run " + r);
						synchronized(retries) {
							if (RunStatus.finishedWithUnexpectedError(r.getStatus())) {
								if (!retries.contains(r.getAlgorithmRunRequest().getHash()) && !RunStatus.isFinished(getStatus()) && getTerminationStatus() == null) {
									AlgorithmRunRequest req = r.getAlgorithmRunRequest();
									req = req.getFreshRequest();
									getRun(req, instid);
									retries.add(req.getHash());
									log.warning("Run " + r.getId() + " finished with unexpected error " + r.getStatus() + "; retrying");
									return;
								} else {
									log.warning("Run " + r.getId() + " failed  with error " + r.getStatus() +"; not retring again.");
								}
							}
							retries.remove(r.getAlgorithmRunRequest().getHash());
						}
						Double val = metric.getEvaluation(r).getValue();
						if (val == null) return;
						Map<Long, List<Double>> m = data.get(impl);
						if (m == null) {
							m = new HashMap<Long, List<Double>>();
							data.put(impl, m);
						}
						List<Double> d = m.get(instid);
						if (d == null) {
							d = new LinkedList<Double>();
							m.put(instid, d);
						}
						d.add(val);
						// only produce logged output every so often
						if (time != null && time - lasttime > outinterval) {
							produceOutput();
							lasttime = time;
						}
					} finally {
						done++;
						inprog--;
						freetime += (maxsubruntime - r.getTotalCpuTime());
						data.notifyAll();
					}
				}
			}
		}
		
		/***
		 * wait for all outstanding runs to finish, or for the run to be terminated
		 */
		private void waitForActiveCompletion() {
			synchronized(data) {
				while (inprog > 0) {
					Double termstat = getTerminationStatus();
					if (termstat != null) finish(termstat);
					try {
						// wait longer if many runs are outstanding, but never more than 5 seconds
						data.wait(Math.min(5000, 100*inprog));
					} catch (InterruptedException ignored) {}
				}
			}
		}
		
		@Override
		protected void start() {
			setStatus(RunStatus.RUNNING);
			try {
				for (int cycle = 0; cycle < runsperinst; cycle++) {
					Long instid = -1L;
					for (final ProblemInstance inst : distribution) {
						// block on seeds
						final long seed = (seedDomain.getSample(rng)).longValue();
						instid++;
						final long iid = instid;
						
						for (final Algorithm target : targets) {
							// don't needlessly run deterministic algorithms multiple times
							if (target.isDeterministic() && cycle > 0) continue;
							
							// be responsive to termination requests
							if (getTerminationStatus() != null) {
								waitForActiveCompletion();
								return;
							}
							
							// end appropriately
							if (done >= maxruns || RunStatus.isFinished(getStatus()) || runsleft == 0) {
								return;
							}
							
							// launch a run
							AlgorithmRunRequest req;
							synchronized(target) {
								target.setProblemInstance(inst);
								if (target.hasScenarioVariable(Semantics.SEED)) target.setScenarioValue(Semantics.SEED, seed);
								if (target.hasScenarioVariable(Semantics.MAX_RUNLENGTH) && maxsubrunlength < Long.MAX_VALUE) target.setScenarioValue(Semantics.MAX_RUNLENGTH, maxsubrunlength);
								if (target.hasScenarioVariable(Semantics.MAX_CPUTIME) && maxsubruntime < Double.MAX_VALUE) target.setScenarioValue(Semantics.MAX_CPUTIME, maxsubruntime);
								req = target.getAlgorithmRunRequest();
							}
							
							getRun(req, combine ? inst.hashCode() : iid);
							
							// wait until there's enough free time to launch another run
							// and also free slots
							synchronized(data) {
								while (maxtime < 1e300 && inprog > 0 && RunStatus.isRunning(getStatus()) &&
										(
												inprog >= maxParallel ||
	//											runner.numFreeSlots() < 1 ||
												freetime <= maxsubruntime )) {
									if (getTotalCpuTime() >= maxtime) {
										terminate(RunStatus.TIMEOUT);
										waitForActiveCompletion();
										return;
									}
									try {
										data.wait(100);
									} catch (InterruptedException ignored) {}
								}
							}
						}
					}
				}
				waitForActiveCompletion();
			} finally {
				finish(getTerminationStatus() == null ? RunStatus.FINISHED : getTerminationStatus());
			}
		}

		private AlgorithmRun getRun(AlgorithmRunRequest req, long instid) {
			synchronized(data) {
				AlgorithmRun run = runner.fetchRun(req);
				instids.put(run, instid);
				run.registerCompletionVisitor(new ComparisonSubrunCompletionVisitor());
				Global.getThreadPool().execute(run);
				runsleft--;
				inprog++;
				freetime -= maxsubruntime;
				return run;
			}
		}
		
		/***
		 * Mark the run as finished, terminate any outstanding runs,
		 * and produce any final output.
		 * @param finalstatus the status of the run after this call.
		 *  This is ignored if the run already has a finished status when the method is called
		 */
		private void finish(Double finalstatus) {
			synchronized(data) {
				setStatus(RunStatus.FINISHING);
				// stop all the still-in-progress runs.
				log.info("finishing with status " + finalstatus);
				if (inprog > 0) {
					runner.terminate(RunStatus.TERMINATED);
					while (inprog > 0) {
						try {
							data.wait();
						} catch (InterruptedException ignored) {}
					}
				}
				waitForActiveCompletion();
				produceOutput(); // produce final output
				setStatus(finalstatus);
			}
		}
		
		@SuppressWarnings("unchecked")
		@Override
		public Map<String, Object> getOutputSnapshot() {
			Map<String, Object> out = new HashMap<String, Object>();
						
			ArrayList<Double> sqs = new ArrayList<Double>(targets.size().intValue()),
				q90s = (ArrayList<Double>) sqs.clone(),
				q75s = (ArrayList<Double>) sqs.clone(),
				q50s = (ArrayList<Double>) sqs.clone(),
				q25s = (ArrayList<Double>) sqs.clone(),
				q10s = (ArrayList<Double>) sqs.clone(),
				means = (ArrayList<Double>) sqs.clone(),
				sds = (ArrayList<Double>) sqs.clone();
			ArrayList<Plot> scds = new ArrayList<Plot>(targets.size().intValue());
			ArrayList<Integer> rls = new ArrayList<Integer>(targets.size().intValue());
			
			Plot p = new Plot();
			boolean hasoutput = false;
			List<List<Double>> friedmanData = new LinkedList<List<Double>>();
			synchronized(data) {
				p.setKey(true);
				p.setKeypos(Plot.LOWERRIGHT);
				p.setXscale(Plot.LOG);
				Collection<Long> univInsts = null;
				
				for (Map<Long, List<Double>> alg: data.values()) {
					if (univInsts == null) univInsts = new HashSet<Long>(alg.keySet());
					else univInsts.retainAll(alg.keySet());
				}
				univInsts = new ArrayList<Long>(univInsts);
				for (ParameterlessAlgorithm alg : targets) {
					// prep other outputs
					DescriptiveStatistics ds = new DescriptiveStatistics();
					Map<Long, List<Double>> dat = data.get(alg);
					rls.add(dat == null ? 0 : dat.size());
					List<Double> fd = new LinkedList<Double>();
					Map<Long, Double> valmap = new HashMap<Long, Double>();
					if (dat != null) for (Map.Entry<Long, List<Double>> pi : dat.entrySet()) {
						Double d = metric.aggregate(pi.getValue());
						if (d != null) {
							ds.addValue(d);
						}
						valmap.put(pi.getKey(), d);
					}
					for (Long l: univInsts) {
						fd.add(valmap.get(l));
					}
					if (ds.getN() > 0) {
						friedmanData.add(fd);
						hasoutput = true;
						List<Number> vals = new LinkedList<Number>();
						for (double d: ds.getValues()) vals.add(d);
						sqs.add(metric.aggregate(vals));
						means.add(ds.getMean());
						sds.add(ds.getStandardDeviation());
						q90s.add(ds.getPercentile(90));
						q75s.add(ds.getPercentile(75));
						q50s.add(ds.getPercentile(50));
						q25s.add(ds.getPercentile(25));
						q10s.add(ds.getPercentile(10));
						Plot pl = Plot.makeSCD(dat, Statistics.MEDIAN);
						scds.add(pl);
						Layer l = new Layer();
						l.setData(pl.getLayer(0).getData());
						l.setLabel(alg.getName());
						l.setType(pl.getLayer(0).getType());
						p.addLayer(l);
					} else {
						sqs.add(null);
						means.add(null);
						sds.add(null);
						q90s.add(null);
						q75s.add(null);
						q50s.add(null);
						q25s.add(null);
						q10s.add(null);
						scds.add(null);
					}
				}
			}
			if (hasoutput) {
				out.put(Semantics.RUNLENGTH, done);
				out.put(SCD_OVERLAY, p);
				out.put(SOLUTION_QUALITIES, sqs);
				out.put(MEANS, means);
				out.put(STDDEVS, sds);
				out.put(Q90S, q90s);
				out.put(Q75S, q75s);
				out.put(Q50S, q50s);
				out.put(Q25S, q25s);
				out.put(Q10S, q10s);
				out.put(RUNLENGTHS, rls);
				out.put(SCDS, scds);
				if (targets.size().intValue() == friedmanData.size() &&
						friedmanData.size() > 1 && friedmanData.get(0).size() > 1) {
					TestResult res = statistics.friedmanTest(friedmanData);
					extraCpuTime += res.getExternalComputeTime();
					out.put(FRIEDMAN_Q, res.getStatistic());
					out.put(FRIEDMAN_P, res.getP());
					if (res.getP() < significance) {
						// do pairwise tests with wilcoxon + bonferroni correction, vs. the best-median-performance algorithm
						double cval = significance / targets.size().doubleValue();
						int minalg = q50s.indexOf(Collections.min(q50s));
						List<Double> bestscores = friedmanData.get(minalg);
//						System.out.println(minalg + " " + q50s);
						List<Double> wilcoxWs = new LinkedList<Double>(), wilcoxPs = new LinkedList<Double>();
						List<Boolean> wilcoxBest = new LinkedList<Boolean>();
//						System.out.println(bestscores);
						for (int i=0; i<friedmanData.size(); i++) {
							if (i == minalg) {
								// no need to test the min vs. itself
								wilcoxBest.add(true);
								wilcoxPs.add(1.0);
								wilcoxWs.add(0.0);
								continue;
							}
							res = statistics.wilcoxonTest(bestscores, friedmanData.get(i), 0.0, Alternative.LESS, true);
//							System.out.println(friedmanData.get(i) + " " + res.getStatistic() + " " + res.getP());
							wilcoxWs.add(res.getStatistic());
							wilcoxPs.add(res.getP());
							wilcoxBest.add(res.getP() > cval);
							extraCpuTime += res.getExternalComputeTime();
						}
						
						out.put(WILCOXON_PS, wilcoxPs);
						out.put(WILCOXON_WS, wilcoxWs);
						out.put(WILCOXON_WINNERS, wilcoxBest);
					}
				}
			}
			return out;
		}
		
		/***
		 * Produce output for logging; uses getOutputSnapshot()
		 */
		private void produceOutput() {
			super.updateOutput(getOutputSnapshot());
		}

		@Override
		public double getFractionCompleted() {
			if (RunStatus.isFinished(getStatus())) return 1.0;
			return Math.max(Math.min(0.99, Math.max(1.0*done/maxruns, getTotalCpuTime() / maxtime)), 0.0);
		}
		
		/**
		 * includes visitors registered to implement a core part of the meta-algorithm,
		 * and any extra cpu time needed for external statistics calculation
		 */
		@Override
		public double measureCpuTime() {
			return super.measureCpuTime() + myVisitorTime + extraCpuTime;
		}
		
		/**
		 * discounts visitors registered to implement a core part of the meta-algorithm
		 */
		@Override
		public double getSubrunOverheadCpuTime() {
			double out = super.getSubrunOverheadCpuTime() - myVisitorTime;
			return out;
		}
	}
}
