package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis;

import java.math.BigInteger;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.logging.Logger;

import net.sf.json.JSONObject;

import org.apache.commons.math.stat.descriptive.DescriptiveStatistics;

import ca.ubc.cs.beta.hal.algorithms.AlgorithmRunRequest;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.MetaAlgorithmImplementation;
import ca.ubc.cs.beta.hal.algorithms.parameters.BooleanDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.IntegerDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ObjectClassDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpace;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpaceBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.RealDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.SampleableDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.analysis.Plot;
import ca.ubc.cs.beta.hal.analysis.Statistics;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.SubrunRunner;
import ca.ubc.cs.beta.hal.problems.InstanceMetricMetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.InstanceDistribution;
import ca.ubc.cs.beta.hal.problems.MetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;
import ca.ubc.cs.beta.hal.utils.ImmutableJsonSerializable;
import ca.ubc.cs.beta.hal.utils.Misc;

import static ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis.AnalysisSemantics.*;

public class SCDBasedAnalysis extends MetaAlgorithmImplementation implements ImmutableJsonSerializable {
			
	private static final String VERSION = "1.1.2";
	private static final Logger log = Logger.getLogger(SCDBasedAnalysis.class.getCanonicalName());
	private static final ParameterSpace DFLT_OUT_SPACE, DFLT_SCN_SPACE, DFLT_CFG_SPACE, SUPP_OPT_SPACE;
	private static final Map<String, String> variables = Misc.asMap(
			SCD, "solution cost distribution",
			MEAN, "mean performance",
			STDDEV, "performance standard deviation",
			Q90, "90th performance percentile",
			Q75, "75th performance percentile",
			Q50, "median performance",
			Q25, "25th performance percentile",
			Q10, "10th performance percentile",
			Semantics.SOLUTION_QUALITY, "solution quality per provided metric",
			MIN_OUTPUT_INTERVAL, "minimum logged output interval (s)",
			COMBINE_IDENTICAL_INSTANCES, "collapse instances with identical hash",
			Semantics.MAX_RUNLENGTH, "maximum number of target algorithm runs",
			MAX_PARALLEL_RUNS, "maximum runs to launch in parallel");
	
	static {
		
		ParameterSpaceBuilder psb = new ParameterSpaceBuilder();
		psb.put(SCD, new ObjectClassDomain(Plot.class));
		psb.put(MEAN, new RealDomain());
		psb.put(STDDEV, new RealDomain());
		psb.put(Q90, new RealDomain());
		psb.put(Q75, new RealDomain());
		psb.put(Q50, new RealDomain());
		psb.put(Q25, new RealDomain());
		psb.put(Q10, new RealDomain());
		psb.put(Semantics.SOLUTION_QUALITY, new RealDomain());
		psb.put(Semantics.RUNLENGTH, Semantics.getDomain(Semantics.RUNLENGTH));
		DFLT_OUT_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		DFLT_CFG_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.SEED, Semantics.getDomain(Semantics.SEED));
		psb.put(Semantics.MAX_CPUTIME, new RealDomain(0.0, null, 1e300));
		psb.put(Semantics.MAX_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(MIN_OUTPUT_INTERVAL, new RealDomain(0., null, 60.0));
		psb.put(COMBINE_IDENTICAL_INSTANCES, new BooleanDomain(false));
		psb.put(MAX_PARALLEL_RUNS, new IntegerDomain(1, null, 100));
		DFLT_SCN_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, new RealDomain(0.0, null, 60.0));
		psb.put(Semantics.MAX_SUBRUN_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(Semantics.RUNS_PER_INSTANCE, new IntegerDomain(1L, null, 1L));
		SUPP_OPT_SPACE = psb.build();
	}
		
	public String describeSemantics(String semantics) {
		String out = variables.get(semantics);
		return out == null ? Semantics.describe(semantics) : out;
	}
	public Collection<String> getDefinedSemantics() {
		return variables.keySet();
	}

	public SCDBasedAnalysis() {
		super(SCDBasedAnalysis.class.getSimpleName(), VERSION, 
				DFLT_CFG_SPACE, DFLT_SCN_SPACE, DFLT_OUT_SPACE, SUPP_OPT_SPACE, 
				Semantics.DETERMINISTIC, false);
	}
	
	@SuppressWarnings("unchecked")
	public Set<Set<String>> getRequiredTags() {
		return Misc.asSet(Misc.asSet(MetaProblemInstance.TAG, MetaProblemInstance.getNumAlgsTag(1),
				InstanceMetricMetaProblemInstance.TAG));
	}

	@Override
	public MetaAlgorithmRun getRun(AlgorithmRunRequest request,
			SubrunRunner runner, Statistics stats) {	
		if (!request.getImplementation().equals(this)) {
			throw new IllegalArgumentException("Request does not correspond to this meta-algorithm implementation");
		}
		return new SingleAlgorithmAnalysisMetaAlgorithmRun(request, runner);
	}
	
	@Override
	public JSONObject buildSpec() {
		JSONObject out = super.buildSpec();
		out.put("version", VERSION);
		return out;
	}

	public static SCDBasedAnalysis fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(SCDBasedAnalysis.class, spec);
		String ver = o.optString("version");
		if (!VERSION.equals(ver) && ver != null && ver.length() > 0) {			
			throw new IllegalArgumentException("Record corresponds to different version of SCDBasedAnalysis plugin; was:" + 
					ver + "; is: " + VERSION );
		}
		return new SCDBasedAnalysis();
	}
	
	private static class SingleAlgorithmAnalysisMetaAlgorithmRun extends MetaAlgorithmRun {

		private volatile double myVisitorTime = 0.0;
		private final ParameterlessAlgorithm target;
		private final InstanceDistribution distribution;
		private final PerformanceMetric metric;
		private final SubrunRunner runner;
		private final Random rng;
		/***
		 * For tracking which instance in the distribution each run corresponds to.
		 * Needed in order to support running on instance distributions with repeated instances,
		 * and to treat each of these repetitions independently
		 */
		private final Map<AlgorithmRun, Long> instids = new HashMap<AlgorithmRun, Long>();
		private final Map<Long, List<Double>> data = new HashMap<Long, List<Double>>();
		private final InstanceMetricMetaProblemInstance instance;
		private final Double maxsubruntime, outinterval;
		private final Long maxruns, runsperinst, maxsubrunlength, maxParallel;
		private volatile double maxtime, freetime, lasttime=0.;
		private volatile long runsleft, inprog, cycle=0L, done=0L;
		private final boolean combine;
		private final SampleableDomain<?> seedDomain;
		
		public SingleAlgorithmAnalysisMetaAlgorithmRun(
				AlgorithmRunRequest request, SubrunRunner runner) {
			super(request, runner);
			this.runner = runner;
			combine = (Boolean) request.getScenarioValue(COMBINE_IDENTICAL_INSTANCES);
			ProblemInstance inst0 = request.getProblemInstance();
			assert inst0 instanceof InstanceMetricMetaProblemInstance;
			
			instance = (InstanceMetricMetaProblemInstance) inst0;
			assert instance.getAlgorithms().size().compareTo(BigInteger.ONE) == 0;
			assert instance.getInstanceDistribution().size() > 0;
			assert instance.getMetric() != null;
			
			maxParallel = ((Number)request.getScenarioValue(MAX_PARALLEL_RUNS)).longValue();
			
			metric = instance.getMetric();
			distribution = instance.getInstanceDistribution();
			if (instance.getAlgorithms() instanceof ParameterizedAlgorithm) {
				log.warning("Got a configrable algorithm, not a grounded algorithm!  Using default configuration...");
				target = ((ParameterizedAlgorithm) instance.getAlgorithms()).getParameterlessAlgorithm();
			} else {
				target = instance.getAlgorithms().get(0);
			}
			
			Long seed = ((Number)request.getScenarioValue(Semantics.SEED)).longValue();
			rng = Global.getRandom(seed);
			
			SampleableDomain<?> d = (SampleableDomain<?>) Semantics.getDomain(Semantics.SEED);
			if (target.hasScenarioVariable(Semantics.SEED)) {
				d = (SampleableDomain<?>) target.getScenarioDomain(Semantics.SEED);
			} 
			seedDomain = d;
			
			Number n;
			n = (Number) request.getScenarioValue(Semantics.MAX_CPUTIME);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(Semantics.MAX_CPUTIME).getDefaultValue();
			maxtime = n.doubleValue();
			freetime = maxtime;
			
			n = (Number) request.getScenarioValue(MIN_OUTPUT_INTERVAL);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(MIN_OUTPUT_INTERVAL).getDefaultValue();
			outinterval = n.doubleValue();
			
			n = (Number) instance.getOption(Semantics.RUNS_PER_INSTANCE);
			if (n == null) n = (Number) SUPP_OPT_SPACE.get(Semantics.RUNS_PER_INSTANCE).getDefaultValue();
			runsperinst = n.longValue();
			
			n = (Number) request.getScenarioValue(Semantics.MAX_RUNLENGTH);
			if (n == null) n = (Number) DFLT_SCN_SPACE.get(Semantics.MAX_RUNLENGTH).getDefaultValue();
			if (distribution.size() > 0 && runsperinst > 0) {
				n = Math.min(n.longValue(), distribution.size()*runsperinst);
			}
			runsleft = n.longValue();
			maxruns = runsleft;
						
			n = (Number) instance.getOption(Semantics.MAX_SUBRUN_CPUTIME);
			if (n != null && n.doubleValue() < Double.MAX_VALUE && target.hasScenarioVariable(Semantics.MAX_CPUTIME)) {
				maxsubruntime = Math.min(n.doubleValue(), maxtime);
			} else {
				maxsubruntime = Double.POSITIVE_INFINITY;
			}
			n = (Number) instance.getOption(Semantics.MAX_RUNLENGTH);
			if (n != null && n.longValue() < Long.MAX_VALUE && target.hasScenarioVariable(Semantics.MAX_RUNLENGTH)) {
				maxsubrunlength = n.longValue();
			} else {
				maxsubrunlength = Long.MAX_VALUE;
			}
		}

		private class AnalysisSubrunCompletionVisitor extends AlgorithmRunVisitor {
			@Override
			public String toString() {
				return "SingleAlgorithmAnalysisCompletionVisitor";
			}
			@Override
			public void reportRunTime(double time) {
				super.reportRunTime(time);
				myVisitorTime += time;
			}
			@Override
			public void visit(AlgorithmRun r) {
				synchronized(data) { // only one update at a time
					try {
						Long instnum;
						synchronized(instids) {
							instnum = instids.remove(r);
						}
						if (instnum == null) throw new RuntimeException("Unexpected run " + r);
						Double val = metric.evaluate(r);
						if (val == null) return;
						List<Double> d = data.get(instnum);
						if (d == null) {
							d = new LinkedList<Double>();
							data.put(instnum, d);
						}
						d.add(val);
						Double time = getTotalCpuTime();
						// only produce logged output every so often
						if (time != null && time - lasttime > outinterval) {
							produceOutput();
							lasttime = time;
							}
					} finally {
						done++;
						inprog--;
						freetime += (maxsubruntime - r.getTotalCpuTime());
						data.notifyAll();
					}
				}
			}
		};	
		
		/*** 
		 * wait for all outstanding runs to finish
		 */
		private void waitForActiveCompletion() {
			Double termstat = getTerminationStatus();
			while (inprog > 0) {
				if (termstat != null) {
					finish(termstat);
					System.out.println("TERM " + termstat);
				}
				synchronized(data) {
					try {
						data.wait(100);
					} catch (InterruptedException e) {
						termstat = getTerminationStatus();
					}
				}
			}
		}
		
		@Override
		protected void start() {
			setStatus(RunStatus.RUNNING);

			Iterator<ProblemInstance> iterator = distribution.iterator();
			
			long instid = 0L;
			
			while (!RunStatus.isFinished(getStatus())) {
				
				// end appropriately when we hit the max number of runs
				if (done >= maxruns || runsleft == 0) {
					while(inprog > 0) {
						waitForActiveCompletion();
					}
					finish(RunStatus.FINISHED);
					continue;
				}
				
				// be responsive to termination requests
				if (getTerminationStatus() != null) {
					finish(getTerminationStatus());
					continue;
				}

				// launch another run
				if (!iterator.hasNext()) {
					if (++cycle < runsperinst) {
						iterator = distribution.iterator();
						instid = 0;
					}
					else {
						runsleft = 0;
						continue;
					}
				}
				AlgorithmRun run;
				ProblemInstance inst = iterator.next();
				synchronized(target) {
					target.setProblemInstance(inst);
					long seed = ((Number)seedDomain.getSample(rng)).longValue(); // always compute it even if we dont use it, to guarantee identical seeds with paired case.
					if (target.hasScenarioVariable(Semantics.SEED)) target.setScenarioValue(Semantics.SEED, seed);
					if (maxsubrunlength < Long.MAX_VALUE) target.setScenarioValue(Semantics.MAX_RUNLENGTH, maxsubrunlength);
					if (maxsubruntime < Double.POSITIVE_INFINITY) target.setScenarioValue(Semantics.MAX_CPUTIME, maxsubruntime);
					run = runner.fetchRun(target.getAlgorithmRunRequest());
				}
				synchronized(instids) {
					instids.put(run, combine ? inst.hashCode() : instid++);
				}
				inprog++;
				run.registerCompletionVisitor(new AnalysisSubrunCompletionVisitor());
				Global.getThreadPool().execute(run);
				runsleft--;
				freetime -= maxsubruntime; 
				// wait until there's enough free time to launch another run
				while (inprog > 0 && RunStatus.isRunning(getStatus()) &&
						(
//								runner.numFreeSlots() < 1 || 
								inprog >= maxParallel ||
								freetime <= maxsubruntime )) {
					if (getTotalCpuTime() >= maxtime) {
						finish(RunStatus.TIMEOUT);
						waitForActiveCompletion();
						return;
					}
					synchronized(data) {
						try {
							if (getTerminationStatus() != null) {
								finish(getTerminationStatus());
							}
							data.wait(100);
						} catch (InterruptedException term) {}
					}
				}
			}
		}

		/***
		 * Mark the run as finished, terminate any outstanding runs, 
		 * and produce any final output.
		 * @param finalstatus the status of the run after this call
		 */
		private void finish(Double finalstatus) {
			synchronized(data) {
				setStatus(RunStatus.FINISHING);
				// stop all the still-in-progress runs.
				log.info("finishing with status " + finalstatus);
				if (inprog > 0) {
					runner.terminate(RunStatus.TERMINATED);
					while (inprog > 0) {
						try {
							data.wait();
						} catch (InterruptedException ignored) {}
					}
				}
				produceOutput(); // produce final output
				setStatus(finalstatus);
			}
		}
				
		@Override
		public Map<String, Object> getOutputSnapshot() {
			Map<String, Object> out = new HashMap<String, Object>();
			DescriptiveStatistics s = new DescriptiveStatistics();
			synchronized(data) {
				for (List<Double> instancePoints : data.values()) {
					s.addValue(metric.aggregate(instancePoints));
				}
				out.put(SCD, Plot.makeSCD(data, Statistics.MEDIAN));
			}
			out.put(Semantics.RUNLENGTH, done);
			
			List<Number> vals = new LinkedList<Number>();
			for (double d: s.getValues()) vals.add(d);
			out.put(Semantics.SOLUTION_QUALITY, metric.aggregate(vals));
			
			out.put(MEAN, s.getMean());
			out.put(STDDEV, s.getStandardDeviation());
			out.put(Q90, s.getPercentile(90));
			out.put(Q75, s.getPercentile(75));
			out.put(Q50, s.getPercentile(50));
			out.put(Q25, s.getPercentile(25));
			out.put(Q10, s.getPercentile(10));
			return out;
		}
		
		/***
		 * Produce output for logging; uses getOutputSnapshot()
		 */
		private void produceOutput() {
			super.updateOutput(getOutputSnapshot());
		}

		@Override
		public double getFractionCompleted() {
			if (RunStatus.isFinished(getStatus())) return 1.0;
			return Math.max(Math.min(0.99, Math.max(1.0*done/maxruns, getTotalCpuTime() / maxtime)), 0.0);		}
		
		/** 
		 * includes visitors registered to implement a core part of the meta-algorithm
		 */
		@Override
		public synchronized double measureCpuTime() {
			return super.measureCpuTime() + myVisitorTime;
		}
		
		/** 
		 * discounts visitors registered to implement a core part of the meta-algorithm
		 */
		@Override
		public double getSubrunOverheadCpuTime() {
			return super.getSubrunOverheadCpuTime() - myVisitorTime;
		}
	}
}
