package ca.ubc.cs.beta.hal.analysis;

import static ca.ubc.cs.beta.hal.analysis.Statistics.Alternative.*;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.json.JSONObject;

import org.apache.commons.io.FileUtils;
import org.rosuda.REngine.REXP;
import org.rosuda.REngine.REXPMismatchException;
import org.rosuda.REngine.REngineException;
import org.rosuda.REngine.RList;
import org.rosuda.REngine.Rserve.RConnection;
import org.rosuda.REngine.Rserve.RserveException;

import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis.AnalysisSemantics;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.executionmanagers.LocalExecutionManager;
import ca.ubc.cs.beta.hal.problems.FileProblemInstance;
import ca.ubc.cs.beta.hal.utils.Global;
import ca.ubc.cs.beta.hal.utils.ImmutableJsonSerializable;
import ca.ubc.cs.beta.hal.utils.Misc;

public class RStatistics extends Statistics implements ImmutableJsonSerializable {
	public static final String RSERVE_HOST = "rServeHost";
	public static final String RSERVE_PORT = "rServePort";
	public static final String R_CMD = "R";
	
	public static final Map<String, String> DFLT_PROPS;
	static {
		Map<String, String> props = new HashMap<String, String>();
		props.put(RSERVE_HOST, "localhost");
		props.put(RSERVE_PORT, "6311");
		props.put(R_CMD, "R");
		DFLT_PROPS = Collections.unmodifiableMap(props);
	}
	private final RConnection c;
	private final ParameterizedAlgorithm Rwilcox;
	private final JsonHelper<RStatistics> helper = new JsonHelper<RStatistics>(this);
	private final Map<String, String> props = new HashMap<String, String>(DFLT_PROPS);
	private final LocalExecutionManager localexec;
	private static final Logger log = Logger.getLogger(RStatistics.class.getCanonicalName());

	private static final Set<String> BCI_SEMANTICS = Collections.unmodifiableSet(
			Misc.asSet(
					AnalysisSemantics.MEAN, AnalysisSemantics.STDDEV,
					AnalysisSemantics.Q10, AnalysisSemantics.Q25,
					AnalysisSemantics.Q50, AnalysisSemantics.Q75,
					AnalysisSemantics.Q90
			)
	);
	
	public RStatistics() {
		this(null);
	}
	
	public RStatistics(Map<String, String> properties) {
		this(properties, false);
	}
	
	public RStatistics(Map<String, String> properties, boolean allowFallback) throws RuntimeException {
		if (properties != null) props.putAll(properties);
		RConnection conn = null;
		LocalExecutionManager xm = null;
		ParameterizedAlgorithm rw = null;
		try {
//			System.err.println("Attempting to connect to Rserve at " + props.get("rServeHost") + ":" + props.get("rServePort"));
			conn = new RConnection(props.get("rServeHost"), Integer.valueOf(props.get("rServePort")));
		} catch (RserveException e) {
			if (!allowFallback) throw new RuntimeException("Could not connect to RServe", e);
			log.warning("Failed to connect to Rserve; " + props.get("rServeHost") + ":" + props.get("rServePort") +"; using command line program " + props.get("rCmd"));
			xm = new LocalExecutionManager();
			String rCmd = props.get("rCmd");
			rw = ParameterizedAlgorithm.fromSpec("{'name':'R', 'path':'.', 'command':'" + rCmd + "', 'deterministic':true, 'inputFormat':{'callstring':['CMD BATCH $infile$ /dev/stdout']},"+
					"'tags':[['R']], " +
					"'inputs':{'infile':{'semantics':'INSTANCE_FILE'}}, " +
					"'outputFormat':{'stdout':['V = $V$, p-value = $P$', 'W = $V$, p-value = $P$', '$confInt$ percent confidence interval:\\n $lowConf$ $highConf$','(pseudo)median\\n $median$']}, " +
					"'outputs':{'V':{'domain':'Real[0, inf)'}, 'P':{'domain':'Real[0,1]'}," +
					"'confInt':{'domain':'Real(0,1)'}, 'lowConf':{'domain':'Real(-inf, inf)'}, 'highConf':{'domain':'Real(-inf, inf)'}, 'median':{'domain':'Real(-inf, inf)'}}}");
		}
		
		
		localexec = xm;
		c = conn;
		Rwilcox = rw;
	}
	
	public boolean isRServe() {
		return c != null;
	}
	
	@Override
	public synchronized TestResult spearmanTest(double[] x, double[] y, Alternative alternative) {
		if (x == null || y == null || x.length != y.length) throw new IllegalArgumentException("x and y must be nonnull equal-length arrays.");
		if (c == null) {
			throw new RuntimeException("Rserve not found");
		}
//		if (c == null) return rCmdSpearmanTest(x, y);
//		return rServeSpearmanTest(x, y);
		StringBuilder script = new StringBuilder("cor.test(x, y,");
		if (alternative == null) alternative = TWO_SIDED;
		if (alternative != GREATER && alternative != LESS && alternative != TWO_SIDED) {
			throw new IllegalArgumentException("Unrecognized alternative; use Statistics.GREATER, Statistics.LESS, or Statistics.TWO_SIDED");
		}
		script.append(" " + (alternative==GREATER ? "\"g\"" : (alternative==LESS ? "\"l\"" : "\"t\"")));
		script.append(", \"spearman\")");
		TestResult out = new TestResult();
		try {
			c.assign("x", x);
			c.assign("y", y);
			RList l = c.eval(script.toString()).asList();
			out.setStatistic(l.at("estimate").asDouble());
			out.setP(l.at("p.value").asDouble());
			// TODO: get confidence interval stuff... not using this at all atm
		} catch (RserveException e) {
			log.warning(e.getMessage());
			if (e.getRequestReturnCode() != 127) throw new RuntimeException(e);
		}catch (Exception e) {
			throw new RuntimeException(e);
		}
		return out;
	}
	
	@Override
	public TestResult wilcoxonTest(double[] listA, double[] listB, Double mu, Alternative alternative, Boolean paired) {
		return rWilcoxonTest(listA, listB, mu, alternative, paired, null, null, null, null);
	}
	
	private TestResult rWilcoxonTest(double[] listA, double[] listB, Double mu, Alternative alternative, Boolean paired, Boolean correct, Boolean exact, Boolean confInt, Double confLevel) {
		if (listA == null) throw new IllegalArgumentException("listA cannot be null");
		paired = paired == null ? getPaired(listA, listB) : paired;
		exact = exact == null ? !getNormal(listA, listB, paired) : exact;
		
		if (c == null) return rCmdWilcoxonTest(listA, listB, mu, alternative, paired, correct, exact, confInt, confLevel);
		return rServeWilcoxonTest(listA, listB, mu, alternative, paired, correct, exact, confInt, confLevel);
	}
	
	private TestResult rCmdWilcoxonTest(double[] listA, double[] listB, Double mu, Alternative alternative, Boolean paired, Boolean correct, Boolean exact, Boolean confInt, Double confLevel) {
		
		TestResult out = new TestResult();
		StringBuilder script = new StringBuilder();
		
		script.append("x <- c(");
		for (Double val: listA) {
			script.append(val == null ? "NA, " : val + ", ");
		}
		script.replace(script.length() - 2, script.length(), ")\n");
		
		if (listB != null) {
			script.append("y <- c(");
			for (Double val: listB) {
				script.append(val == null ? "NA, " : val + ", ");
			}
			script.replace(script.length() - 2, script.length(), ")\n");
		}
		
		script.append("wilcox.test(x, y");
		if (listB == null) script.append(" = NULL");
		if (mu != null) script.append(", mu = " + mu);
		if (alternative != null) {
			if (alternative != GREATER && alternative != LESS && alternative != TWO_SIDED) {
				throw new IllegalArgumentException("Unrecognized alternative; use Statistics.GREATER, Statistics.LESS, or Statistics.TWO_SIDED");
			}
			script.append(", alternative = " + (alternative==GREATER ? "\"g\"" : (alternative==LESS ? "\"l\"" : "\"t\"")));
		}
		if (paired != null) script.append(", paired = " + (paired ? "TRUE" : "FALSE"));
		if (correct != null) script.append(", correct = " + (correct ? "TRUE" : "FALSE"));
		if (exact != null) script.append(", exact = " + (exact ? "TRUE" : "FALSE"));
		if (confInt != null) {
			script.append(", conf.int = " + (confInt ? "TRUE" : "FALSE"));
			if (confInt && confLevel != null) {
				if (confLevel <= 0 || confLevel >= 1) throw new IllegalArgumentException("confLevel must be between 0 and 1");
				script.append(", conf.level = " + confLevel);
			}
		}
		script.append(")\nq(runLast=FALSE)\n"); // otherwise it takes like 1s/run just to quit
		
		FileProblemInstance instance;
		File tempfile;
		try {
			tempfile = File.createTempFile("wilcox", ".R");
			tempfile.deleteOnExit();
			FileUtils.writeStringToFile(tempfile, script.toString());
			instance = new FileProblemInstance(tempfile);
			instance.addTags("R");
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		ParameterizedAlgorithm wr = Rwilcox.clone();
		wr.setProblemInstance(instance);
		AlgorithmRun r = null;
		try {
			r = localexec.fetchRun(wr.getAlgorithmRunRequest(), null);
			Global.getThreadPool().execute(r);
			r.waitFor();
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		} finally {
			tempfile.delete();
		}
		out.setHighConf((Double)r.getLastOutputValueOnly("highConf"));
		out.setLowConf((Double)r.getLastOutputValueOnly("lowConf"));
		out.setConfInt((Double)r.getLastOutputValueOnly("confInt"));
		if (out.getConfInt() != null) out.setConfInt(out.getConfInt()/100);
		out.setP((Double)r.getLastOutputValueOnly("P"));
		out.setStatistic((Double)r.getLastOutputValueOnly("V"));
		out.setComputeTime(r.getTotalCpuTime());
		
		return out;
	}
	
	private synchronized TestResult rServeWilcoxonTest(double[] listA, double[] listB, Double mu, Alternative alternative, Boolean paired, Boolean correct, Boolean exact, Boolean confInt, Double confLevel) {

		TestResult out = new TestResult();
		StringBuilder script = new StringBuilder();
		script.append("wilcox.test(x, y");
		if (listB == null) script.append(" = NULL");
		if (mu != null) script.append(", mu = " + mu);
		if (alternative != null) {
			script.append(", alternative = " + (alternative==GREATER ? "\"g\"" : (alternative==LESS ? "\"l\"" : "\"t\"")));
		}
		if (paired != null) script.append(", paired = " + (paired ? "TRUE" : "FALSE"));
		if (correct != null) script.append(", correct = " + (correct ? "TRUE" : "FALSE"));
		if (exact != null) script.append(", exact = " + (exact ? "TRUE" : "FALSE"));
		if (confInt != null) {
			script.append(", conf.int = " + (confInt ? "TRUE" : "FALSE"));
			if (confInt && confLevel != null) {
				if (confLevel <= 0 || confLevel >= 1) throw new IllegalArgumentException("confLevel must be between 0 and 1");
				script.append(", conf.level = " + confLevel);
			}
		}
		script.append(")");
//		System.out.println(script);
		
		try {
			c.assign("x", listA);
			c.assign("y", listB);
			RList l = c.eval(script.toString()).asList();
//			System.out.println(Arrays.toString(listA) + " " + Arrays.toString(listB) + " " + l.at("p.value").asString() + " " + l.at("statistic").asString());
			out.setStatistic(l.at("statistic").asDouble());
			out.setP(l.at("p.value").asDouble());
			// TODO: get confidence interval stuff... not using this at all atm
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		return out;
	}
	
	public synchronized TestResult friedmanTest(double[][] data) {

		TestResult out = new TestResult();
		double[] longdata = new double[data.length*data[0].length];
		for (int i=0; i<data.length; i++) {
			for (int j=0; j<data[i].length; j++) {
				longdata[i*data[i].length + j] = data[i][j];
			}
		}
		try {
			c.assign("x", longdata);
			c.voidEval("d <- matrix(x, byrow=1, nrow=" + data.length + ")");
			RList l = c.eval("friedman.test(d)").asList();
			out.setStatistic(l.at("statistic").asDouble());
			out.setP(l.at("p.value").asDouble());
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		return out;
	}
	
	private Boolean permtestSupported;
	public synchronized boolean isPermutationTestSupported() {
		if (permtestSupported != null) return permtestSupported;
		
		if (!isRServe()) permtestSupported = false;
		else {
			// Call 'is.element("perm", installed.packages()[,1]) to check if package perm is installed.
			try {
				REXP eval = c.eval("is.element('perm', installed.packages()[,1])");
				permtestSupported = (eval.asInteger() != 0);
			} catch (RserveException e) {
				log.log(Level.SEVERE, "", e);
				permtestSupported = false;
			} catch (REXPMismatchException e) {
				throw new RuntimeException("Code should never reach this point.", e);
			}
		}
		if (!permtestSupported) {
			log.warning("Permutation test not supported for this R-statistics: 'perm' must be an installed library.");
		}
		return permtestSupported;
	}

	/**
	 * Requires Rserve and package 'perm' installed. Permutation test for 2 means.
	 * 
	 * @param listA
	 * @param listB
	 * @param alternative
	 *            may be null
	 * @param exact
	 *            may be null
	 * @return
	 */
	@Override
	public synchronized TestResult permutationTest(double[] listA, double[] listB, Alternative alternative, Boolean exact) {
		StringBuilder script = new StringBuilder();
		TestResult result = new TestResult();
		
		/*
		 * See: http://rgm2.lab.nig.ac.jp/RGM2/func.php?rd_id=perm:perm
		 */
		script.append("permTS(x, y");
		if (alternative != null) {
			char alt = 't';
			switch (alternative) {
			case GREATER: alt = 'g'; break;
			case LESS:    alt = 'l'; break;
			}
			script.append(", alternative='").append(alt).append("'");
		}
		if (exact != null) {
			script.append(", exact=").append(exact ? "'T'" : "'F'");
		}
		script.append(")");
		
		try {
			c.voidEval("library(perm)");
			c.assign("x", listA);
			c.assign("y", listB);
		
			RList l = c.eval(script.toString()).asList();
			//== [statistic, estimate, parameter, p.value, null.value, alternative, method, data.name, p.values, p.conf.int, nmc]
			log.info("Permutation p: " + l.at("p.value").asDouble());
			
			result.setStatistic(l.at("estimate").asDouble());
			result.setP(l.at("p.value").asDouble());
			
		} catch (RserveException e) {
			log.warning(e.getMessage());
			if (e.getRequestReturnCode() != 127) throw new RuntimeException(e);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		
		return result;
	}

	/**
	 * Requires Rserve and package 'perm' installed. Permutation test for 2
	 * means.
	 * 
	 * @param listA
	 * @param listB
	 * @param alternative
	 *            may be null
	 * @param method
	 *            may be null
	 * @return
	 */
	public synchronized TestResult permutationTest(double[] listA, double[] listB, Alternative alternative, PermutationTestMethod method) {
		StringBuilder script = new StringBuilder();
		TestResult result = new TestResult();
		
		/*
		 * See: http://rgm2.lab.nig.ac.jp/RGM2/func.php?rd_id=perm:perm
		 */
		script.append("permTS(x, y");
		
		if (alternative != null) {
			char alt = 't';
			switch (alternative) {
			case GREATER: alt = 'g'; break;
			case LESS:    alt = 'l'; break;
			}
			script.append(", alternative='").append(alt).append("'");
		}
		if (method != null) {
			String methodName;
			switch (method) {
				case PCLT:
					methodName = "pclt"; break;
				case EXACT_COMPLETE_ENUMERATION:
					methodName = "exact.ce"; break;
				case EXACT_MONTE_CARLO:
					methodName = "exact.mc"; break;
				case EXACT_NETWORK:
					methodName = "exact.net"; break;
				default:
					throw new IllegalArgumentException("Unsupported method: " + method);
			}
			script.append(", method='").append(methodName).append("'");
		}
		script.append(")");
		
		try {
			c.voidEval("library(perm)");
			c.assign("x", listA);
			c.assign("y", listB);
		
			RList l = c.eval(script.toString()).asList();
			//== [statistic, estimate, parameter, p.value, null.value, alternative, method, data.name, p.values, p.conf.int, nmc]
			
			log.info("Permutation p: " + l.at("p.value").asDouble());
			
			result.setStatistic(l.at("estimate").asDouble());
			result.setP(l.at("p.value").asDouble());
			
		} catch (RserveException e) {
			log.warning(e.getMessage());
			if (e.getRequestReturnCode() != 127) throw new RuntimeException(e);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		
		return result;
	}

	@Override
	public Set<String> getSupportedBootstrapCISemantics() {
		return new HashSet<String>(BCI_SEMANTICS);
	}

	//===== Notes from Holger
	//	At long last, here are my notes on doing bootstrapping in R:
	//	R supports bootstrap analyses, including BCa conf intv
	//
	//	# bootstrapping + confidence intervals
	//	library(boot)
	//	fd = read.table("averna-f.dat")
	//	b = boot(fd[,1], samplemean, R=1000)
	//	boot.ci(b, conf = c(0.90,0.95),type = c("norm","basic","perc","stud","bca"))
	//
	//	notes:
	//	- on Nima's RNA data, BCa doesn't compute (no bias?), but on sample RTD
	//	data, it works
	//	and gives rather different CIs from std percentile intv
	//	- m-out-of-n (with m smaller n) sampling (as opposed to m-out-of-m)
	//	seems to solve many problems, but does R support it
	//	- bootstrap for hyp testing exists, but must be careful about null-hyp;
	//	best to use permutation test when applicable
	//
	//	I've attached averna-f.dat, so you can see the format I used for input.
	//	I suggest you use the 95-percentile bootstrap interval ("perc")
	//	- for that, it is sufficient to use
	//
	//	boot.ci(b, conf = c(0.95),type = c("perc")) instead of the last two
	//	lines above.
	//
	//	Let me know if you have any questions!
	//
	//	Holger
	public synchronized ConfidenceIntervals bootstrapCI(double[] data, int rep, double conf, Set<String> semantics) {
		ConfidenceIntervals intervals = new ConfidenceIntervals();
		Set<String> toCompute = new HashSet<String>(semantics);
		ArrayList<String> computedSemantics = new ArrayList<String>();
		
		for (String s: toCompute) {
			if (!BCI_SEMANTICS.contains(s)) {
				log.warning("Removing unsupported semantic: " + s);
				toCompute.remove(s);
			}
		}
		
		if (toCompute.isEmpty()) {
			log.warning("Evaluating empty set semantics.");
			return intervals;
		}
		

		List<String> fStrings = new LinkedList<String>();
		StringBuilder stats = new StringBuilder("stats <- function(x,y) c(");
		
		//== For each output, add it to the things to evaluate. Then build the function string from fStrings.
		//== First, handle mean and sd.
		if (toCompute.contains(AnalysisSemantics.MEAN)) {
			fStrings.add("mean=mean(x[y])");
			computedSemantics.add(AnalysisSemantics.MEAN);
		}
		if (toCompute.contains(AnalysisSemantics.STDDEV)) {
			fStrings.add("sd=sd(x[y])");
			computedSemantics.add(AnalysisSemantics.STDDEV);
		}
		
		//== Next, handle quantiles.
		List<String> quantileStrings = new LinkedList<String>();
		String[] quantileStringPairs = new String[] {
				AnalysisSemantics.Q10, ".10",
				AnalysisSemantics.Q25, ".25",
				AnalysisSemantics.Q50, ".50",
				AnalysisSemantics.Q75, ".75",
				AnalysisSemantics.Q90, ".90"
		};
		
		for (int i = 0; i < quantileStringPairs.length; i+=2) {
			String semantic = quantileStringPairs[i];
			if (toCompute.contains(semantic)) {
				computedSemantics.add(semantic);
				quantileStrings.add(quantileStringPairs[i+1]);
			}
		}

		if (!quantileStrings.isEmpty()) {
			StringBuilder sb = new StringBuilder("q=quantile(x[y], probs=c(");
			Iterator<String> it = quantileStrings.iterator();

			sb.append(it.next());
			while (it.hasNext()) sb.append(",").append(it.next());
			sb.append("), type=6)"); //type=6 corresponds to DescriptiveStatistics

			fStrings.add(sb.toString());
		}
				
		//=== Build the statistics string from the gathered parts.
		Iterator<String> it = fStrings.iterator();
		stats.append(it.next());
		while (it.hasNext()) stats.append(", ").append(it.next());
		stats.append(")");
		
		String statistic = stats.toString();
		log.info(statistic);
		
		//=== Now, set up the R function call.
		try {
			c.voidEval("library(boot)");
			c.voidEval("samplemean <- function(x,y) mean(x[y])"); // #sum(x[y])/length(y)
			c.voidEval(statistic);
			c.assign("fd", data);
		} catch (RserveException e) {
			throw new RuntimeException(e);
		} catch (REngineException e) {
			throw new RuntimeException(e);
		}
		
		//=== Call the function and get results for each output.
		try {
			c.voidEval("b = boot(fd, stats, R=" + rep + ")");
			
			RList l;
			int index = 1;
			
			for (String semantic : computedSemantics) {
				try {
					l = c.eval("boot.ci(b, conf=" + String.valueOf(conf)
							+ ", type ='perc', index=" + index + ")").asList();
					double[] r = l.at("percent").asDoubles();
					if (r == null || r.length < 5) {
						log.warning("Missing result for " +  semantic + ": " + Arrays.toString(r));
					}
					ConfidenceInterval ci = new ConfidenceInterval(r[3], r[4], conf);
					intervals.put(semantic, ci);
					
				} catch (Exception e) {
					log.log(Level.SEVERE, "Error evaluating " + semantic, e);
				}
				index++;
			}
		} catch (RserveException e) {
			throw new RuntimeException(e);
		}
		
		return intervals;
	}
	
	public String toSpec() {
		return helper.getSpec();
	}

	public String getHash() {
		return helper.getHash();
	}
	
	@Override
	public int hashCode() {
		return helper.getHashCode();
	}
	
	@Override
	public boolean equals(Object o) {
		return helper.getEquals(o);
	}

	public JSONObject buildSpec() {
		JSONObject spec = helper.getStubSpec();
		spec.put("properties", props);
		return spec;
	}
	
	public static RStatistics fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(RStatistics.class, spec).getJSONObject("properties");
		Map<String, String> properties = new HashMap<String, String>();
		for (Object k: o.keySet()) {
			properties.put((String)k, o.getString((String)k));
		}
		return new RStatistics(properties);
	}

	@Override
	public String toFullSpec() {
		return toSpec();
	}
}
