package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.LinkedList;
import java.util.logging.Level;

import org.apache.commons.io.FileUtils;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import ca.ubc.cs.beta.hal.algorithms.Algorithm;
import ca.ubc.cs.beta.hal.algorithms.AlgorithmOutputValue;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithmList;
import ca.ubc.cs.beta.hal.algorithms.MetaAlgorithmImplementation;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSetting;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSettingBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.analysis.GnuplotPlotter;
import ca.ubc.cs.beta.hal.analysis.Plot;
import ca.ubc.cs.beta.hal.analysis.Plotter;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun.RunStatus;
import ca.ubc.cs.beta.hal.environments.Environment;
import ca.ubc.cs.beta.hal.environments.ExecutionManager;
import ca.ubc.cs.beta.hal.environments.ExecutionManager.OutputHandlingOption;
import ca.ubc.cs.beta.hal.environments.FullAccessDataManager;
import ca.ubc.cs.beta.hal.environments.datamanagers.InvalidURIException;
import ca.ubc.cs.beta.hal.environments.datamanagers.SQLiteDataManager;
import ca.ubc.cs.beta.hal.environments.datamanagers.sql.AbstractSQLDataManager;
import ca.ubc.cs.beta.hal.environments.executionmanagers.LocalExecutionManager;
import ca.ubc.cs.beta.hal.environments.executionmanagers.SSHExecutionManager;
import ca.ubc.cs.beta.hal.problems.*;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;

import static ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis.AnalysisSemantics.*;


public class SCDBasedAnalysisTest {

	private static Algorithm analyser;
	private static Level l;
	private static FullAccessDataManager dm0, dm;
	
	@Before
	public void clear() {
		dm.clear();
	}
	
	@BeforeClass
	public static void setup() throws InvalidURIException {
		dm = new SQLiteDataManager(URI.create("jdbc:sqlite:test/singcmp.db"));
		l = Global.getBaseLogger().getLevel();
		dm0 = Global.getDataManager();
		Global.getBaseLogger().setLevel(Level.WARNING);
		analyser = getTestAlgorithm();
		Global.setDataManager(dm);
	}
	
	@AfterClass
	public static void shutdown() {
		dm = null;
		Global.setDataManager(dm0);
		Global.getBaseLogger().setLevel(l);
	}
	
	public static Algorithm getTestAlgorithm() {
		ParameterSettingBuilder psb = new ParameterSettingBuilder();
		psb.put(Semantics.RUNS_PER_INSTANCE, 2);
		return getTestAlgorithm(psb.build());
	}
	
	public static Algorithm getTestAlgorithm(ParameterSetting instOptions) {
		// get the meta-algorithm
		MetaAlgorithmImplementation alg = new SCDBasedAnalysis();
		Algorithm analyser = alg.getDefaultParameterizedAlgorithm();
		
		String json = "{'command':'java', 'tags':[['dummy']], 'outputFormat':{'stdout':['foroutput $v$ $t$']}, 'outputs':{'v':{'domain':'Real()'}, 't':{'domain':'Real()'}}, 'name':'json1', 'deterministic':false, 'inputFormat':{'callstring':['-classpath bin:.:*:lib/*:../hal/bin:../HAL/bin ca.ubc.cs.beta.hal.utils.TimeWaster $time$ [$hitime$]']}, 'inputs':{'seed':{'domain':'Integer[1, inf)', 'default':0.5 ,'semantics':'SEED'}, 'time':{'domain':'Real[0, inf)', 'default':0.5}, 'hitime':{'domain':'Real[0, inf)', 'default':0}}}";
		
		ParameterizedAlgorithm alg1 = ParameterizedAlgorithm.fromSpec(json);
		alg1.setConfigurationValue("time", 0.5);
		alg1.setConfigurationValue("hitime", 0.5);
				
		ParameterlessAlgorithmList algs = new ParameterlessAlgorithmList(alg1.getParameterlessAlgorithm());
		
		PerformanceMetric metric = PerformanceMetric.PAR10;
		
		// create some dummy instances for our dummy algorithm
		LinkedList<ProblemInstance> il = new LinkedList<ProblemInstance>();
		try {
			File id = new File(Global.getReferencePoint(), "tstinsts"+System.currentTimeMillis());
			id.deleteOnExit();
			FileUtils.forceMkdir(id);
			
			for (int i=0; i<3; i++) {
				File f = File.createTempFile("dummy", ".inst", id);
				FileUtils.writeStringToFile(f, f.getName()); // they need to have different hashes or the DB knows they're the same
				f.deleteOnExit();
				FileProblemInstance pi = new FileProblemInstance(f);
				il.add(pi);
				il.add(pi); // 2 copies of every instance
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		InstanceList instances = new InstanceList("dummylist", il);
		instances.addTagsToAllInstances(Tag.DUMMY, Semantics.EXPORTABLE);
		
		
		MetaProblemInstance inst = new InstanceMetricMetaProblemInstance(algs, instances, metric, "dummyAnalysisProb", instOptions);
//		inst.setOption(Semantics.MAX_SUBRUN_CPUTIME, 0.5);
				
		analyser.setProblemInstance(inst);
//		analyser.setScenarioValue(Semantics.MAX_CPUTIME, 5);
//		analyser.setScenarioValue(Semantics.MAX_RUNLENGTH, 1000);
		analyser.setScenarioValue("minOutputInterval", 1);
		return analyser;
		
	}
	
	@Test
	public void testRunsLocal() throws Exception {
		
		LocalExecutionManager xm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		
		final Environment env = new Environment("local unittest", xm, Environment.Preference.MAX_SIMULT_TARGET_RUNS, 2);

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
		Global.getThreadPool().execute(run);

		while (!RunStatus.isFinished(run.getStatus())) {
			System.out.println(run.getFractionCompleted());
			Thread.sleep(500);
		}
		System.out.println(run.getFractionCompleted());
		
		run.waitFor();
		assertTrue(RunStatus.isFinished(run.getStatus()));
		
		int targinsts = ((InstanceMetricMetaProblemInstance)run.getProblemInstance()).getInstanceDistribution().size();
		assertEquals(targinsts, 
				((Plot)run.getLastOutputValueOnly(SCD)).getLayer(0).getData().size());
		
		
		run.waitForVisitors();
		
		while(((AbstractSQLDataManager)dm).getActiveConnections() > 0) {
			Thread.sleep(500);
			System.err.println("WAITING");
		}
		
		assertEquals(run.getSubrunCount(), ((Number)run.getLastOutputValueOnly(Semantics.RUNLENGTH)).longValue());
		
		System.out.println(run.getOutput());
		System.out.println("total " + run.getTotalCpuTime());
		System.out.println("measured " + run.getMeasuredCpuTime());
		System.out.println("overhead " + run.getOverheadCpuTime());
		System.out.println("subs " + run.getSubrunTotalCpuTime());
		System.out.println("subM " + run.getSubrunMeasuredCpuTime());
		System.out.println("subO " + run.getSubrunOverheadCpuTime());


		
//		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
//		for (AlgorithmOutputValue kv: run.getOutput(SCD)) {
//			try {
//				p.display((Plot)kv.getValue());
//			} catch (Exception ignored){}
//		}
	}
	
//	@Test
	public void testTerminatesOK() throws Exception {
		LocalExecutionManager xm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		
		final Environment env = new Environment("local unittest", xm, Environment.Preference.MAX_SIMULT_TARGET_RUNS, 2);

		for (int i=0; i<100; i++) {
			AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
			Global.getThreadPool().execute(run);
	
			Thread.sleep(1000+Global.getRandom().nextInt(3000));
			run.terminate();
						
			run.waitForVisitors();
			assertTrue(RunStatus.isFinished(run.getStatus()));
			
			System.out.println("Done " + i);
			clear();
		}
	}
	
	@Test
	public void testRunsLocalCombineInsts() throws Exception {
		
		LocalExecutionManager xm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		
		analyser.setScenarioValue(COMBINE_IDENTICAL_INSTANCES, true);
		final Environment env = new Environment("local unittest", xm, Environment.Preference.MAX_SIMULT_TARGET_RUNS, 2);

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
		Global.getThreadPool().execute(run);

		run.waitFor();
		assertTrue(RunStatus.isFinished(run.getStatus()));
		
		int targinsts = ((InstanceMetricMetaProblemInstance)run.getProblemInstance()).getInstanceDistribution().size();
		assertEquals(targinsts/2, 
				((Plot)run.getLastOutputValueOnly(SCD)).getLayer(0).getData().size());
		run.waitForVisitors();
		
		while(((AbstractSQLDataManager)dm).getActiveConnections() > 0) {
			Thread.sleep(500);
			System.err.println("WAITING");
		}
		
		System.out.println(run.getOutput());
		analyser.setScenarioValue(COMBINE_IDENTICAL_INSTANCES, false);

		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
		for (AlgorithmOutputValue kv: run.getOutput(SCD)) {
			try {
				p.display((Plot)kv.getValue());
			} catch (Exception ignored){}
		}
	}
	
	@Test
	public void testRunsLocalTerminate() throws Exception {
		
		LocalExecutionManager xm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		
		final Environment env = new Environment("local unittest", xm, Environment.Preference.MAX_SIMULT_TARGET_RUNS, 10);

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
		Global.getThreadPool().execute(run);
		double time = 0.8;
		Thread.sleep((long)(1000*time));
		run.terminate();
		run.waitForVisitors();
		double stat = run.getStatus();
		assertTrue(RunStatus.isFinished(stat));
		assertTrue(RunStatus.isTerminated(stat));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(run.getSubrunMeasuredCpuTime() + run.getOverheadCpuTime() + run.getVisitorCpuTime(), run.getTotalCpuTime(), 0.5*run.getTotalCpuTime());
		assertEquals(time, run.getTotalCpuTime(), time*.9); // just needs to be in the right ballpark, and nonzero
	}
	
	@Test
	public void testRunsLocalNewProcess() throws Exception {
		
		LocalExecutionManager xm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		
		final Environment env = new Environment("local unittest", xm, dm);

		analyser.getAlgorithmRunRequest().setForceNewProcess(true);
		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
		Global.getThreadPool().execute(run);

//		System.out.println("HN " + run.getHostName());
//		System.out.println(run + " " + ((ConnectionPooledSQLDataManager)dm).getActiveConnections());
//		while (!RunStatus.isFinished(run.getStatus())) {
//			System.out.println("* " + run.getStatus() + " " + run.getCpuTime() + " " + run.getLastOutputValueOnly() + " " + ((ConnectionPooledSQLDataManager)dm).getActiveConnections());
//			Thread.sleep(1000);
//		}
//		System.out.println(run.getStatus() + " " + run.getCpuTime() + " " + run.getOutput());
		
		run.waitForVisitors();
		
		assertTrue(RunStatus.isFinished(run.getStatus()));
//		assertTrue(RunStatus.timedOut(run.getStatus()));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(run.getSubrunMeasuredCpuTime() + run.getOverheadCpuTime() + run.getVisitorCpuTime(), run.getTotalCpuTime(), 0.5*run.getTotalCpuTime());
		assertEquals(((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue(), run.getTotalCpuTime(), 
				ExecutionManager.ADD_TIME_BUFFER + ((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue()*(ExecutionManager.MULT_TIME_BUFFER-1));
		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);		
		Plot plot = (Plot)run.getLastOutputValueOnly(SCD);
		p.display(plot);
	}
	
	@Test
	public void testRunsSSHLocalTarg() throws Exception {

		// the first meta-algorithm happens over ssh; everything else locally after that.
		LocalExecutionManager lm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		SSHExecutionManager xm = new SSHExecutionManager(URI.create("ssh://localhost/" + new File(".").getAbsolutePath()), OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		final Environment env = new Environment("local unittest", xm, lm, lm, dm, null);

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
		Global.getThreadPool().execute(run);

		run.waitFor();
		assertTrue(RunStatus.isFinished(run.getStatus()));
//		assertTrue(RunStatus.timedOut(run.getStatus()));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(run.getSubrunMeasuredCpuTime() + run.getOverheadCpuTime() + run.getVisitorCpuTime(), run.getTotalCpuTime(), 0.5*run.getTotalCpuTime());
		assertEquals(((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue(), run.getTotalCpuTime(), 
				ExecutionManager.ADD_TIME_BUFFER + ((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue()*(ExecutionManager.MULT_TIME_BUFFER-1));

				
		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
		p.display((Plot)run.getLastOutputValueOnly(SCD));
	}
	
	@Test
	public void testRunsLocalNewProcTerm() throws Exception {
				
		// the first meta-algorithm happens over ssh; everything else locally after that.
		LocalExecutionManager lm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		final Environment env = new Environment("local unittest", lm);

		analyser.getAlgorithmRunRequest().setForceNewProcess(true);
		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());
		Global.getThreadPool().execute(run);

		Thread.sleep(1500); // takes a while for the process to get going b/c jvm
		run.terminate();
		run.waitFor();
		assertTrue(RunStatus.isFinished(run.getStatus()));
		assertTrue(RunStatus.terminated(run.getStatus()));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(run.getSubrunMeasuredCpuTime() + run.getOverheadCpuTime() + run.getVisitorCpuTime(), run.getTotalCpuTime(), 0.5*run.getTotalCpuTime());
		assertEquals(0.7, run.getTotalCpuTime(), 0.7);
	}
	
	@Test
	public void testRunsLocalSSHTarg() throws Exception {
		
		// the first meta-algorithm happens over ssh; everything else locally after that.
		LocalExecutionManager lm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		SSHExecutionManager xm = new SSHExecutionManager(URI.create("ssh://localhost/" + new File(".").getAbsolutePath()), OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		final Environment env = new Environment("local unittest", lm, xm, xm, dm, null);

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());	
		Global.getThreadPool().execute(run);

		run.waitFor();
		assertTrue(RunStatus.isFinished(run.getStatus()));
		assertTrue(RunStatus.timedOut(run.getStatus()));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(run.getSubrunMeasuredCpuTime() + run.getOverheadCpuTime() + run.getVisitorCpuTime(), run.getTotalCpuTime(), 0.5*run.getTotalCpuTime());
		assertEquals(((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue(), run.getTotalCpuTime(), 
				ExecutionManager.ADD_TIME_BUFFER + ((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue()*(ExecutionManager.MULT_TIME_BUFFER-1));

				
		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
		p.display((Plot)run.getLastOutputValueOnly(SCD));
	}
	
	@Test
	public void testRunsSSHSSHTarg() throws Exception {
		
		// every run involves a new connection over ssh
		SSHExecutionManager xm = new SSHExecutionManager(URI.create("ssh://localhost/" + new File(".").getAbsolutePath()), OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		final Environment env = new Environment("local unittest", xm, dm);

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());
		Global.getThreadPool().execute(run);

		run.waitFor();
		assertTrue(RunStatus.isFinished(run.getStatus()));
		assertTrue(RunStatus.timedOut(run.getStatus()));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(run.getSubrunMeasuredCpuTime() + run.getOverheadCpuTime() + run.getVisitorCpuTime(), run.getTotalCpuTime(), 0.5*run.getTotalCpuTime());
		assertEquals(((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue(), run.getTotalCpuTime(), 
				ExecutionManager.ADD_TIME_BUFFER + ((Number) analyser.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue()*(ExecutionManager.MULT_TIME_BUFFER-1));
				
		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
		p.display((Plot)run.getLastOutputValueOnly(SCD));
	}
	
}
